import pytest

from sqlfmt.analyzer import Analyzer
from sqlfmt.comment import Comment
from sqlfmt.exception import SqlfmtBracketError, SqlfmtParsingError
from sqlfmt.mode import Mode
from sqlfmt.tokens import Token, TokenType
from tests.util import read_test_data


def test_simple_query_parsing(all_output_modes: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_analyzer/test_simple_query_parsing.sql"
    )

    q = all_output_modes.dialect.initialize_analyzer(
        all_output_modes.line_length
    ).parse_query(source_string=source_string)

    assert q
    assert q.source_string == source_string
    assert len(q.lines) == 6

    expected_line_depths = [0, 1, 1, 1, 0, 0]
    actual_line_depths = [line.depth[0] for line in q.lines]
    assert actual_line_depths == expected_line_depths

    assert q.nodes

    assert len(q.tokens) == 26
    assert isinstance(q.tokens[0], Token)

    expected_tokens = [
        Token(type=TokenType.UNTERM_KEYWORD, prefix="", token="select", spos=0, epos=6),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=6, epos=7),
        Token(
            type=TokenType.NAME,
            prefix="    ",
            token="a_long_field_name",
            spos=7,
            epos=28,
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=28, epos=29),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=29, epos=30),
        Token(
            type=TokenType.NAME,
            prefix="    ",
            token="another_long_field_name",
            spos=30,
            epos=57,
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=57, epos=58),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=58, epos=59),
        Token(type=TokenType.BRACKET_OPEN, prefix="    ", token="(", spos=59, epos=64),
        Token(type=TokenType.NAME, prefix="", token="one_field", spos=64, epos=73),
        Token(type=TokenType.OPERATOR, prefix=" ", token="+", spos=73, epos=75),
        Token(type=TokenType.NAME, prefix=" ", token="another_field", spos=75, epos=89),
        Token(type=TokenType.BRACKET_CLOSE, prefix="", token=")", spos=89, epos=90),
        Token(type=TokenType.WORD_OPERATOR, prefix=" ", token="as", spos=90, epos=93),
        Token(type=TokenType.NAME, prefix=" ", token="c", spos=93, epos=95),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=95, epos=96),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="from", spos=96, epos=100
        ),
        Token(type=TokenType.NAME, prefix=" ", token="my_schema", spos=100, epos=110),
        Token(type=TokenType.DOT, prefix="", token=".", spos=110, epos=111),
        Token(
            type=TokenType.QUOTED_NAME,
            prefix="",
            token='"my_QUOTED_ table!"',
            spos=111,
            epos=130,
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=130, epos=131),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="where", spos=131, epos=136
        ),
        Token(type=TokenType.NAME, prefix=" ", token="one_field", spos=136, epos=146),
        Token(type=TokenType.OPERATOR, prefix=" ", token="<", spos=146, epos=148),
        Token(
            type=TokenType.NAME, prefix=" ", token="another_field", spos=148, epos=162
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=162, epos=162),
    ]

    assert q.tokens == expected_tokens


def test_parsing_error(default_analyzer: Analyzer) -> None:
    source_string = "select $"
    with pytest.raises(SqlfmtParsingError):
        _ = default_analyzer.parse_query(source_string=source_string)


def test_case_statement_parsing(default_analyzer: Analyzer) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_analyzer/test_case_statement_parsing.sql"
    )

    q = default_analyzer.parse_query(source_string=source_string)

    assert q
    assert q.source_string == source_string
    assert len(q.lines) == 20

    expected_line_depths = [0, 1, 1, 1, 2, 2, 1, 1, 2, 2, 2, 3, 1, 1, 2, 1, 1, 1, 1, 0]

    computed_line_depths = [line.depth[0] for line in q.lines]
    assert computed_line_depths == expected_line_depths

    # there are 6 case statements in the test data
    assert len([t for t in q.tokens if t.type is TokenType.STATEMENT_START]) == 6
    assert len([t for t in q.tokens if t.type is TokenType.STATEMENT_END]) == 6


def test_cte_parsing(default_analyzer: Analyzer) -> None:
    source_string, _ = read_test_data("unit_tests/test_analyzer/test_cte_parsing.sql")

    q = default_analyzer.parse_query(source_string=source_string)

    assert q
    assert q.source_string == source_string
    assert len(q.lines) == 3

    expected_line_depths = [0, 1, 0]

    computed_line_depths = [line.depth[0] for line in q.lines]
    assert computed_line_depths == expected_line_depths

    expected_node_depths = [
        0,  # with
        1,  # \n
        1,  # my_cte
        1,  # as
        1,  # (
        2,  # select
        3,  # 1
        3,  # ,
        3,  # b
        3,  # ,
        3,  # another_field
        2,  # from
        3,  # my_schema
        3,  # .
        3,  # my_table
        1,  # )
        1,  # \n
        0,  # select
        1,  # *
        0,  # from
        1,  # my_cte
        1,  # \n
    ]

    computed_node_depths = [node.depth[0] for node in q.nodes]
    assert computed_node_depths == expected_node_depths


def test_multiline_parsing(default_analyzer: Analyzer) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_analyzer/test_multiline_parsing.sql"
    )

    q = default_analyzer.parse_query(source_string=source_string)

    assert q
    assert q.source_string == source_string
    assert len(q.lines) < len(source_string.split("\n"))

    assert TokenType.COMMENT_START not in [token.type for token in q.tokens]
    assert TokenType.COMMENT_END not in [token.type for token in q.tokens]

    expected_comments = [
        Comment(
            token=Token(
                type=TokenType.COMMENT,
                prefix="",
                token=(
                    "/*\n * This is a typical multiline comment.\n"
                    " * It contains newlines.\n"
                    " * And even /* some {% special characters %}\n"
                    " * but we're not going to parse those\n*/"
                ),
                spos=157,
                epos=310,
            ),
            is_standalone=True,
            previous_node=q.nodes[2],
        ),
        Comment(
            token=Token(
                type=TokenType.COMMENT,
                prefix=" ",
                token=(
                    "/* This is a multiline comment in very bad style,\n"
                    "    * which starts and ends on lines with other tokens.\n    */"
                ),
                spos=386,
                epos=499,
            ),
            is_standalone=True,
            previous_node=q.nodes[17],
        ),
        Comment(
            token=Token(
                type=TokenType.COMMENT,
                prefix="",
                token=(
                    "{#\n # And this is a nice multiline jinja comment\n"
                    " # that we will also handle.\n#}"
                ),
                spos=757,
                epos=837,
            ),
            is_standalone=True,
            previous_node=q.nodes[38],
        ),
        Comment(
            token=Token(
                type=TokenType.COMMENT,
                prefix=" ",
                token="/* what!?! */",
                spos=860,
                epos=874,
            ),
            is_standalone=False,
            previous_node=q.nodes[42],
        ),
    ]

    expected_content = [
        Token(
            type=TokenType.JINJA_EXPRESSION,
            prefix="",
            token=(
                "{{\n    config(\n        materialized='table',\n        sort='id',\n"
                "        dist='all',\n"
                "        post_hook='grant select on {{ this }} to role bi_role'\n"
                "    )\n}}"
            ),
            spos=0,
            epos=155,
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=155, epos=156),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=156, epos=157),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="with", spos=312, epos=316
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=316, epos=317),
        Token(type=TokenType.NAME, prefix="    ", token="source", spos=317, epos=327),
        Token(type=TokenType.WORD_OPERATOR, prefix=" ", token="as", spos=327, epos=330),
        Token(type=TokenType.BRACKET_OPEN, prefix=" ", token="(", spos=330, epos=332),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="select", spos=332, epos=338
        ),
        Token(type=TokenType.STAR, prefix=" ", token="*", spos=338, epos=340),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix=" ", token="from", spos=340, epos=345
        ),
        Token(
            type=TokenType.JINJA_EXPRESSION,
            prefix=" ",
            token="{{ ref('my_model') }}",
            spos=345,
            epos=367,
        ),
        Token(type=TokenType.BRACKET_CLOSE, prefix="", token=")", spos=367, epos=368),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=368, epos=369),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=369, epos=370),
        Token(type=TokenType.NAME, prefix="    ", token="renamed", spos=370, epos=381),
        Token(type=TokenType.WORD_OPERATOR, prefix=" ", token="as", spos=381, epos=384),
        Token(type=TokenType.BRACKET_OPEN, prefix=" ", token="(", spos=384, epos=386),
        Token(
            type=TokenType.UNTERM_KEYWORD,
            prefix="  ",
            token="select",
            spos=499,
            epos=507,
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=507, epos=508),
        Token(
            type=TokenType.NAME, prefix="            ", token="id", spos=508, epos=522
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=522, epos=523),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=523, epos=524),
        Token(
            type=TokenType.NAME,
            prefix="            ",
            token="another_field",
            spos=524,
            epos=549,
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=549, epos=550),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=550, epos=551),
        Token(
            type=TokenType.NAME,
            prefix="            ",
            token="and_another",
            spos=551,
            epos=574,
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=574, epos=575),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=575, epos=576),
        Token(
            type=TokenType.NAME,
            prefix="            ",
            token="and_still_another",
            spos=576,
            epos=605,
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=605, epos=606),
        Token(
            type=TokenType.UNTERM_KEYWORD,
            prefix="        ",
            token="from",
            spos=606,
            epos=618,
        ),
        Token(type=TokenType.NAME, prefix=" ", token="source", spos=618, epos=625),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=625, epos=626),
        Token(
            type=TokenType.BRACKET_CLOSE, prefix="    ", token=")", spos=626, epos=631
        ),
        Token(type=TokenType.COMMA, prefix="", token=",", spos=631, epos=632),
        Token(
            type=TokenType.JINJA_STATEMENT,
            prefix=" ",
            token=(
                '{% set my_variable_in_bad_style = [\n        "a",\n'
                '        "short",\n        "list",\n        "of",\n'
                '        "strings"\n    ] %}'
            ),
            spos=632,
            epos=755,
        ),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=755, epos=756),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=756, epos=757),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix="", token="select", spos=839, epos=845
        ),
        Token(type=TokenType.STAR, prefix=" ", token="*", spos=845, epos=847),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix=" ", token="from", spos=847, epos=852
        ),
        Token(type=TokenType.NAME, prefix=" ", token="renamed", spos=852, epos=860),
        Token(
            type=TokenType.UNTERM_KEYWORD, prefix=" ", token="where", spos=874, epos=880
        ),
        Token(type=TokenType.NAME, prefix=" ", token="true", spos=880, epos=885),
        Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=885, epos=885),
    ]

    actual_comments = []
    for line in q.lines:
        actual_comments.extend(line.comments)
    assert actual_comments == expected_comments
    assert q.tokens == expected_content


def test_star_parsing(default_analyzer: Analyzer) -> None:
    space_star = "select * from my_table\n"
    space_star_q = default_analyzer.parse_query(source_string=space_star)

    assert space_star_q
    assert len(space_star_q.nodes) == 5
    assert space_star_q.nodes[1].prefix == " ", (
        "There should be a space between select and star in select *"
    )

    dot_star = "select my_table.* from my_table\n"
    dot_star_q = default_analyzer.parse_query(source_string=dot_star)

    assert dot_star_q
    assert len(dot_star_q.nodes) == 7
    assert dot_star_q.nodes[3].prefix == "", (
        "There should be no space between dot and star in my_table.*"
    )


@pytest.mark.parametrize(
    "source, expected_prefix",
    [
        ("select sum(1)", ""),
        ("over (partition by abc)", " "),
        ("with cte as (select 1)", " "),
        ("select 1 + (1-3)", " "),
        ("where something in (select id from t)", " "),
    ],
)
def test_open_paren_parsing(
    source: str, expected_prefix: str, default_analyzer: Analyzer
) -> None:
    q = default_analyzer.parse_query(source_string=source)

    assert q
    for node in q.nodes:
        if node.token.token == "(":
            assert node.prefix == expected_prefix, (
                "Open paren prefixed by wrong number of spaces"
            )


def test_unterminated_multiline_token(default_analyzer: Analyzer) -> None:
    source_string = "{% \n config = {}\n"

    with pytest.raises(SqlfmtBracketError) as excinfo:
        _ = default_analyzer.parse_query(source_string=source_string)

    assert "Unterminated multiline" in str(excinfo.value)


def test_unmatched_bracket_error(default_analyzer: Analyzer) -> None:
    source_string = "select case )\n"
    with pytest.raises(SqlfmtBracketError) as excinfo:
        _ = default_analyzer.parse_query(source_string=source_string)

    assert "Closing bracket ')'" in str(excinfo.value)


def test_case_insensitive_bracket_matching(default_analyzer: Analyzer) -> None:
    source_string = "--fmt: off\n select CASE when true then 1 end\n"
    q = default_analyzer.parse_query(source_string=source_string)
    assert q.nodes[-2].token.type is TokenType.STATEMENT_END


@pytest.mark.parametrize(
    "source_string",
    [
        "select ( end\n",
        "select end\n",
        "end\n",
    ],
)
def test_no_raise_bracket_error_on_end_name(
    default_analyzer: Analyzer, source_string: str
) -> None:
    q = default_analyzer.parse_query(source_string=source_string)
    assert "end" in str(q)


def test_get_rule(default_analyzer: Analyzer) -> None:
    for rule in default_analyzer.rules:
        matched_rule = default_analyzer.get_rule(rule.name)
        assert matched_rule == rule

    with pytest.raises(ValueError):
        _ = default_analyzer.get_rule("bar")


def test_match_first_jinja_Tag(default_analyzer: Analyzer) -> None:
    """
    Make sure we lazily match only the first tag
    """
    source_string = (
        "{{ event_cte.source_cte_name}}.{{ event_cte.primary_key }} "
        "|| '-' || '{{ event_cte.event_name }}'"
    )
    q = default_analyzer.parse_query(source_string)
    assert len(q.lines) == 1
    assert str(q.nodes[0]).strip() == "{{ event_cte.source_cte_name}}"


def test_jinja_block_parsing(default_analyzer: Analyzer) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_analyzer/test_jinja_block_parsing.sql"
    )
    q = default_analyzer.parse_query(source_string)
    assert q
    assert len(q.lines) == 22
    types = [node.token.type for node in q.nodes if not node.is_newline]
    expected_types = [
        TokenType.JINJA_BLOCK_START,  # {% macro ...(contents, num_times) %}
        TokenType.JINJA_BLOCK_START,  # {% if contents == "foo" %}
        TokenType.JINJA_BLOCK_START,  # {%- for _ in range(num_times * 10) %}
        TokenType.JINJA_EXPRESSION,  # {{ contents }}
        TokenType.JINJA_BLOCK_END,  # {% endfor %}
        TokenType.JINJA_BLOCK_KEYWORD,  # {% elif contents == "bar" %}
        TokenType.JINJA_BLOCK_START,  # {% if num_times > 10 %}
        TokenType.JINJA_BLOCK_START,  # {%- for _ in range(num_times * 5) %}
        TokenType.QUOTED_NAME,  # "TIMES 5!!"
        TokenType.JINJA_EXPRESSION,  # {{ contents }}
        TokenType.JINJA_BLOCK_END,  # {% endfor %}
        TokenType.JINJA_BLOCK_KEYWORD,  # {% else %}
        TokenType.JINJA_BLOCK_START,  # {%- for _ in range(num_times * 2) %}
        TokenType.JINJA_EXPRESSION,  # {{ contents }}
        TokenType.JINJA_BLOCK_END,  # {% endfor %}
        TokenType.JINJA_BLOCK_END,  # {% endif %}
        TokenType.JINJA_BLOCK_KEYWORD,  # {% else %}
        TokenType.JINJA_BLOCK_START,  # {%- for _ in range(num_times) %}
        TokenType.JINJA_EXPRESSION,  # {{ contents }}
        TokenType.JINJA_BLOCK_END,  # {% endfor %}
        TokenType.JINJA_BLOCK_END,  # {% endif %}
        TokenType.JINJA_BLOCK_END,  # {% endmacro %}
    ]
    assert types == expected_types


@pytest.mark.parametrize(
    "source_string",
    [
        "create TABLE a (\n    A text --test; it's a comment\n)\n",
        "create table a (x int); -- trailing; comment\n",
        "create table a (\n    x int  # hash; comment\n)\n",
        "create table a (\n    x int  /* block; comment */\n)\n",
    ],
)
def test_semicolon_in_unsupported_ddl_comment(
    default_analyzer: Analyzer, source_string: str
) -> None:
    """
    A semicolon inside a comment must not terminate an unsupported statement.
    See https://github.com/tconbeer/sqlfmt/issues/838
    """
    q = default_analyzer.parse_query(source_string=source_string)
    comment_tokens = [
        comment.token.token for line in q.lines for comment in line.comments
    ]
    assert any(";" in token for token in comment_tokens)



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import CREATE_CLONABLE, group
from sqlfmt.rules.core import CORE
from sqlfmt.tokens import TokenType

CLONE = [
    *CORE,
    Rule(
        name="unterm_keyword",
        priority=1300,
        pattern=group(CREATE_CLONABLE, r"clone") + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
    Rule(
        name="word_operator",
        priority=1500,
        pattern=group(
            r"at",
            r"before",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.WORD_OPERATOR
            ),
        ),
    ),
]



import shutil
from pathlib import Path
from typing import Iterable, Iterator, List, Tuple, Union

TEST_DIR = Path(__file__).parent
BASE_DIR = TEST_DIR / "data"
RESULTS_DIR = TEST_DIR / ".results"


def read_test_data(relpath: Union[Path, str]) -> Tuple[str, str]:
    """reads a test file contents and returns a tuple of strings corresponding to
    the unformatted and formatted examples in the test file. If the test file doesn't
    include a ')))))__SQLFMT_OUTPUT__(((((' sentinel, returns the same string twice
    (as the input is assumed to be pre-formatted). relpath is relative to
    tests/data/"""
    SENTINEL = ")))))__SQLFMT_OUTPUT__((((("

    test_path = BASE_DIR / relpath

    with open(test_path, "r") as test_file:
        lines = test_file.readlines()

    source_query: List[str] = []
    formatted_query: List[str] = []

    target = source_query

    for line in lines:
        if line.rstrip() == SENTINEL:
            target = formatted_query
            continue
        target.append(line)

    if source_query and not formatted_query:
        formatted_query = source_query[:]

    return "".join(source_query).strip() + "\n", "".join(formatted_query)


def _safe_create_results_dir() -> Path:
    results_dir = RESULTS_DIR
    results_dir.mkdir(exist_ok=True)
    return results_dir


def delete_results_dir() -> None:
    shutil.rmtree(RESULTS_DIR, ignore_errors=True)


def check_formatting(expected: str, actual: str, ctx: str = "") -> None:
    try:
        assert expected == actual, (
            "Formatting error. Output file written to tests/.results/"
        )
    except AssertionError as e:
        import inspect

        results_dir = _safe_create_results_dir()

        caller = inspect.stack()[1].function
        if ctx:
            caller += "-"
            ctx = ctx.replace("/", "-")

        if ctx.endswith(".sql"):
            suffix = ""
        else:
            suffix = ".sql"

        p = results_dir / (caller + ctx + suffix)
        with open(p, "w") as f:
            f.write(actual)
        raise e


def discover_test_files(relpaths: Iterable[Union[str, Path]]) -> Iterator[Path]:
    for p in [BASE_DIR / p for p in relpaths]:
        if p.is_file() and p.suffix == ".sql":
            yield p
        elif p.is_dir():
            yield from (discover_test_files(p.iterdir()))


def copy_test_data_to_tmp(relpaths: List[str], tmp_path: Path) -> Path:
    """
    Reads in test data from an existing file or directory, and creates a new file
    at the temp_path with the source query from the original test data file.

    Returns the path to the temporary file
    """

    for abspath in discover_test_files(relpaths):
        file_contents, _ = read_test_data(abspath)

        with open(tmp_path / abspath.name, "w") as tmp_file:
            tmp_file.write(file_contents)

    return tmp_path


def copy_config_file_to_dst(file_name: str, dst_path: Path) -> Path:
    CONFIG_DIR = BASE_DIR / "config"
    file_path = CONFIG_DIR / file_name
    assert file_path.is_file()

    new_file_path = dst_path / "pyproject.toml"
    shutil.copyfile(file_path, new_file_path)
    return new_file_path



from dataclasses import dataclass
from typing import List, Tuple

from sqlfmt.comment import Comment
from sqlfmt.line import Line
from sqlfmt.node import Node
from sqlfmt.node_manager import NodeManager


@dataclass
class LineSplitter:
    node_manager: NodeManager

    def maybe_split(self, line: Line) -> List[Line]:
        """
        Evaluates a line for splitting. If line matches criteria for splitting,
        returns a list of new lines; otherwise returns a list of only the original line.

        We used to do this recursively, but very long lines (with >500 splits) would
        raise RecursionError.
        """

        if line.formatting_disabled:
            return [line]

        new_lines: List[Line] = []
        comments = line.comments
        head = 0
        always_split_after = never_split_after = False
        for i, node in enumerate(line.nodes):
            if node.is_newline:
                if head == 0:
                    new_lines.append(line)
                else:
                    new_line, comments = self.split_at_index(
                        line, head, i, comments, no_tail=True
                    )
                    assert not comments, (
                        "Comments must be empty here or we'll drop them"
                    )
                    new_lines.append(new_line)
                return new_lines
            elif (
                i > head
                and not never_split_after
                and not node.formatting_disabled
                and (always_split_after or self.maybe_split_before(node))
            ):
                new_line, comments = self.split_at_index(line, head, i, comments)
                new_lines.append(new_line)
                head = i
                # node now follows a new newline node, so we need to update
                # its previous node (this can impact its depth)
                node.previous_node = new_line.nodes[-1]

            always_split_after, never_split_after = self.maybe_split_after(node)

        new_line, comments = self.split_at_index(line, head, -1, comments, no_tail=True)
        assert not comments, "Comments must be empty here or we'll drop them"
        new_lines.append(new_line)
        return new_lines

    def maybe_split_before(self, node: Node) -> bool:
        """
        Return True if we should split before node
        """
        if (
            # if there is a multiline node on this line and it isn't the
            # only thing on this line, then split before the multiline node
            node.is_multiline_jinja
            # always split before any unterm kw
            or node.is_unterm_keyword
            # always split before any opening jinja block
            or node.is_opening_jinja_block
            # always split before operators
            or node.is_operator
            # always split before any node that decreases depth
            or node.is_closing_bracket
            or node.is_closing_jinja_block
            # always split before a node that divides queries
            or node.divides_queries
        ):
            return True
        # split if an opening bracket immediately follows
        # a closing bracket
        elif self.maybe_split_between_brackets(node):
            return True
        else:
            return False

    def maybe_split_between_brackets(self, node: Node) -> bool:
        """
        Return true if this is an open bracket that follows
        a closing bracket. This is typically for BQ
        array indexing, like split(my_field)[offset(1)],
        or dictionary accessing, like my_json['outer']['inner']
        """
        return (
            node.is_opening_bracket
            and node.previous_node is not None
            and node.previous_node.is_closing_bracket
        )

    def maybe_split_after(self, node: Node) -> Tuple[bool, bool]:
        """
        Return True, False if we should always split after node
        Retrun False, True if we should never split after node
        Return False, False if splitting after should depend on the
        contents of the next node
        """
        if (
            # always split after any comma that doesn't end a line
            node.is_comma
            # always split after a token that increases depth
            or node.is_opening_bracket
            or node.is_opening_jinja_block
            or node.is_unterm_keyword
            # always split after a token that divides queries
            or node.divides_queries
        ):
            return True, False
        elif node.formatting_disabled:
            return False, True
        else:
            return False, False

    def split_at_index(
        self,
        line: Line,
        head: int,
        index: int,
        comments: List[Comment],
        no_tail: bool = False,
    ) -> Tuple[Line, List[Comment]]:
        """
        Return a new line comprised of the nodes line[head:index], plus a newline node.

        Also return a list of comments that need to be included in the tail
        of the split line. This includes inline comments and if the head
        is just a comma, all comments.
        """
        if index == -1:
            new_nodes = line.nodes[head:]
        else:
            assert index > head, "Cannot split at start of line!"
            new_nodes = line.nodes[head:index]

        assert new_nodes, "Cannot split a line without nodes!"

        if no_tail:
            head_comments, tail_comments = comments, []

        elif comments:
            if new_nodes[0].is_comma:
                head_comments, tail_comments = [], comments
            elif index == len(line.nodes) - 2 and line.nodes[index].is_operator:
                # the only thing after the split is an operator + \n, so keep the
                # comments with the stuff before the operator
                head_comments, tail_comments = comments, []
            else:
                head_comments, tail_comments = [], []
                for comment in comments:
                    if (
                        comment.is_standalone
                        or comment.is_multiline
                        or comment.previous_node is None
                    ):
                        head_comments.append(comment)
                    elif comment.is_inline:
                        tail_comments.append(comment)
                    elif comment.is_c_style and comment.previous_node in new_nodes:
                        head_comments.append(comment)
                    else:
                        tail_comments.append(comment)

        else:  # no comments
            head_comments, tail_comments = [], []

        new_line = Line.from_nodes(
            previous_node=new_nodes[0].previous_node,
            nodes=new_nodes,
            comments=head_comments,
        )
        if not new_line.nodes[-1].is_newline:
            self.node_manager.append_newline(new_line)

        return new_line, tail_comments



import pickle
from importlib import metadata
from pathlib import Path
from typing import Dict, Iterable, List, Tuple

from platformdirs import user_cache_dir

from sqlfmt.mode import Mode
from sqlfmt.report import STDIN_PATH, SqlFormatResult

Cache = Dict[Path, Tuple[float, int]]


def get_cache_file() -> Path:
    """
    Returns the path to the cache file on disk
    """
    sqlfmt_version = metadata.version("shandy-sqlfmt")
    cache_dir = Path(user_cache_dir(appname="sqlfmt"))
    cache_file = cache_dir / f"cache-{sqlfmt_version}.pickle"
    return cache_file


def load_cache() -> Cache:
    """
    Returns a Cache (a dictionary keyed by file path) by loading
    from a pickle saved to disk
    """
    cache_file = get_cache_file()
    try:
        with cache_file.open("rb") as f:
            cache: Cache = pickle.load(f)
            return cache
    except (
        pickle.UnpicklingError,
        ValueError,
        IndexError,
        FileNotFoundError,
        ModuleNotFoundError,
    ):
        return {}


def check_cache(cache: Cache, p: Path) -> bool:
    """
    Returns True if the path is in the cache and the cached stats match the
    file on disk
    """
    if p == STDIN_PATH or not cache:
        return False
    else:
        path = p.resolve()
        if path in cache:
            return _get_cache_info(path) == cache[path]
        else:
            return False


def write_cache(cache: Cache, results: List[SqlFormatResult], mode: Mode) -> None:
    """
    Updates cache with results, then dumps cache to disk
    """
    cache_file = get_cache_file()
    cache_file.parent.mkdir(parents=True, exist_ok=True)
    new_cache = cache.copy()
    for path in _gen_cache_keys_for_updates(results, mode):
        updated_info = _get_cache_info(path)
        new_cache[path] = updated_info
    with open(cache_file, "wb") as f:
        pickle.dump(new_cache, f)


def clear_cache() -> None:
    """
    Deletes the cache file on disk, if it exists
    """
    p = get_cache_file()
    try:
        p.unlink()
    except FileNotFoundError:
        pass


def _get_cache_info(path: Path) -> Tuple[float, int]:
    """
    Returns a tuple of (modified_time, file_size) for the path; this tuple is
    persisted to the cache, and we check the files on disk against this cached
    value to determine if we need to format the file again
    """
    stat = path.resolve().stat()
    file_info = (stat.st_mtime, stat.st_size)
    return file_info


def _gen_cache_keys_for_updates(
    results: Iterable[SqlFormatResult], mode: Mode
) -> Iterable[Path]:
    """
    Takes an interable of SqlfmtResults and yields paths to files that should
    be updated in the cache, based on the result of the sqlfmt run
    """
    gen = (
        res.source_path.resolve()
        for res in results
        if res.source_path
        and res.source_path != STDIN_PATH
        and _should_update_cache(res, mode)
    )
    yield from gen


def _should_update_cache(result: SqlFormatResult, mode: Mode) -> bool:
    """
    Takes a single SqlfmtResult and returns True if that result indicates that
    the associated file should be updated in the cache
    """
    if result.has_error or result.from_cache:
        return False
    elif not result.has_changed:
        return True
    elif mode.check or mode.diff:
        return False
    else:
        return True



import re

import pytest

from sqlfmt import actions
from sqlfmt.analyzer import Analyzer
from sqlfmt.exception import SqlfmtBracketError, StopRulesetLexing
from sqlfmt.rules import FUNCTION, JINJA
from sqlfmt.tokens import Token, TokenType
from tests.util import read_test_data


@pytest.fixture
def jinja_analyzer(default_analyzer: Analyzer) -> Analyzer:
    default_analyzer.push_rules(JINJA)
    return default_analyzer


@pytest.fixture
def function_analyzer(default_analyzer: Analyzer) -> Analyzer:
    default_analyzer.push_rules(FUNCTION)
    return default_analyzer


def test_raise_sqlfmt_bracket_error(default_analyzer: Analyzer) -> None:
    with pytest.raises(SqlfmtBracketError):
        # we need a match, but any match will do.
        s = ")"
        match = re.match(r"(\))", s)
        assert match
        actions.raise_sqlfmt_bracket_error(default_analyzer, "", match)


def test_add_node_to_buffer(default_analyzer: Analyzer) -> None:
    source_string = "select a, b, c\n"
    rule = default_analyzer.get_rule("unterm_keyword")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_node_to_buffer(
        default_analyzer, source_string, match, TokenType.UNTERM_KEYWORD
    )

    assert default_analyzer.comment_buffer == []
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.node_buffer) == 1

    node = default_analyzer.node_buffer[0]
    assert node.is_unterm_keyword

    assert default_analyzer.pos == 6

    rule = default_analyzer.get_rule("name")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_node_to_buffer(default_analyzer, source_string, match, TokenType.NAME)

    assert default_analyzer.comment_buffer == []
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.node_buffer) == 2

    node = default_analyzer.node_buffer[1]
    assert node.token.type is TokenType.NAME

    assert default_analyzer.pos == 8


def test_safe_add_node_to_buffer(default_analyzer: Analyzer) -> None:
    source_string = "end\n"
    rule = default_analyzer.get_rule("statement_end")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    with pytest.raises(SqlfmtBracketError):
        actions.add_node_to_buffer(
            default_analyzer, source_string, match, TokenType.STATEMENT_END
        )
    assert default_analyzer.node_buffer == []
    # does not raise
    actions.safe_add_node_to_buffer(
        default_analyzer, source_string, match, TokenType.STATEMENT_END, TokenType.NAME
    )

    assert default_analyzer.comment_buffer == []
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.node_buffer) == 1

    node = default_analyzer.node_buffer[0]
    assert node.token.type is TokenType.NAME

    assert default_analyzer.pos == 3


def test_add_comment_to_buffer(default_analyzer: Analyzer) -> None:
    source_string = "-- a comment\n"
    rule = default_analyzer.get_rule("comment")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_comment_to_buffer(default_analyzer, source_string, match)

    assert default_analyzer.node_buffer == []
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.comment_buffer) == 1

    comment = default_analyzer.comment_buffer[0]
    assert comment.token.token == source_string.strip()

    assert default_analyzer.pos == 12


def test_handle_newline_with_nodes(default_analyzer: Analyzer) -> None:
    source_string = "a_name\n"
    rule = default_analyzer.get_rule("name")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_node_to_buffer(default_analyzer, source_string, match, TokenType.NAME)
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.node_buffer) > 0
    assert default_analyzer.comment_buffer == []

    rule = default_analyzer.get_rule("newline")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.handle_newline(default_analyzer, source_string, match)
    assert default_analyzer.pos == len(source_string)
    assert len(default_analyzer.line_buffer) == 1
    assert default_analyzer.node_buffer == []
    assert default_analyzer.comment_buffer == []
    line = default_analyzer.line_buffer[0]
    assert str(line) == "a_name\n"


def test_handle_newline_empty(default_analyzer: Analyzer) -> None:
    source_string = "\n\n\n"
    assert default_analyzer.line_buffer == []
    assert default_analyzer.comment_buffer == []
    assert default_analyzer.node_buffer == []
    assert default_analyzer.pos == 0
    rule = default_analyzer.get_rule("newline")
    for i in range(1, 4):
        match = rule.program.match(source_string, default_analyzer.pos)
        assert match
        actions.handle_newline(default_analyzer, source_string, match)
        assert default_analyzer.pos == i
        assert len(default_analyzer.line_buffer) == i
        assert default_analyzer.node_buffer == []
        assert default_analyzer.comment_buffer == []
        line = default_analyzer.line_buffer[i - 1]
        assert str(line) == "\n"


def test_handle_newline_leading_comments(default_analyzer: Analyzer) -> None:
    source_string = "-- a comment\n\nsomething\n"
    assert default_analyzer.line_buffer == []
    assert default_analyzer.comment_buffer == []
    assert default_analyzer.node_buffer == []

    comment_rule = default_analyzer.get_rule("comment")
    match = comment_rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_comment_to_buffer(default_analyzer, source_string, match)
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.comment_buffer) == 1
    assert default_analyzer.node_buffer == []

    nl_rule = default_analyzer.get_rule("newline")
    match = nl_rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.handle_newline(default_analyzer, source_string, match)
    # do NOT append a line yet
    assert default_analyzer.line_buffer == []
    assert default_analyzer.node_buffer == []
    assert len(default_analyzer.comment_buffer) == 1

    match = nl_rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.handle_newline(default_analyzer, source_string, match)
    # STILL do NOT append a line yet
    assert default_analyzer.line_buffer == []
    assert default_analyzer.node_buffer == []
    assert len(default_analyzer.comment_buffer) == 1

    name_rule = default_analyzer.get_rule("name")
    match = name_rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.add_node_to_buffer(default_analyzer, source_string, match, TokenType.NAME)
    assert default_analyzer.line_buffer == []
    assert len(default_analyzer.node_buffer) == 1
    assert len(default_analyzer.comment_buffer) == 1

    match = nl_rule.program.match(source_string, pos=default_analyzer.pos)
    assert match
    actions.handle_newline(default_analyzer, source_string, match)
    # now we finally create a line
    assert len(default_analyzer.line_buffer) == 1
    assert default_analyzer.node_buffer == []
    assert default_analyzer.comment_buffer == []

    line = default_analyzer.line_buffer[0]
    assert len(line.comments) == 1
    assert len(line.nodes) == 2  # includes nl node


@pytest.mark.parametrize(
    "source_string,has_preceding_star,has_preceding_newline,expected_type",
    [
        (" except", True, False, TokenType.WORD_OPERATOR),
        (" except", True, True, TokenType.WORD_OPERATOR),
        (" except", False, False, TokenType.SET_OPERATOR),
        (" except", False, True, TokenType.SET_OPERATOR),
        (" except all", True, False, TokenType.SET_OPERATOR),  # this is a syntax error
        (" union all", False, False, TokenType.SET_OPERATOR),
    ],
)
def test_handle_set_operator(
    default_analyzer: Analyzer,
    source_string: str,
    has_preceding_star: bool,
    has_preceding_newline: bool,
    expected_type: TokenType,
) -> None:
    rule = default_analyzer.get_rule("set_operator")
    match = rule.program.match(source_string, pos=default_analyzer.pos)
    assert match

    if has_preceding_star:
        t = Token(type=TokenType.STAR, prefix="", token="*", spos=0, epos=1)
        n = default_analyzer.node_manager.create_node(t, previous_node=None)
        default_analyzer.node_buffer.append(n)
    elif has_preceding_newline:
        t_nl = Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=0, epos=1)
        n_nl = default_analyzer.node_manager.create_node(t_nl, previous_node=None)
        default_analyzer.node_buffer.append(n_nl)

    if has_preceding_newline and has_preceding_star:
        t_nl = Token(type=TokenType.NEWLINE, prefix="", token="\n", spos=1, epos=2)
        n_nl = default_analyzer.node_manager.create_node(t_nl, previous_node=n)
        default_analyzer.node_buffer.append(n_nl)

    actions.handle_set_operator(default_analyzer, source_string, match)

    node = default_analyzer.node_buffer[-1]
    assert node.token.type == expected_type


@pytest.mark.parametrize(
    "source_string,start_name,end_name",
    [
        (
            "{% config(field='{{this}}') %}",
            "jinja_statement_start",
            "jinja_statement_end",
        ),
        ("{%- set var = that -%}", "jinja_statement_start", "jinja_statement_end"),
        ("{{ simple_var }}", "jinja_expression_start", "jinja_expression_end"),
        ("{{ macro(arg1, arg2) }}", "jinja_expression_start", "jinja_expression_end"),
    ],
)
def test_handle_jinja(
    jinja_analyzer: Analyzer,
    source_string: str,
    start_name: str,
    end_name: str,
) -> None:
    if "%" in source_string:
        token_type = TokenType.JINJA_STATEMENT
    else:
        token_type = TokenType.JINJA_EXPRESSION
    start_rule = jinja_analyzer.get_rule(start_name)
    match = start_rule.program.match(source_string)
    assert match, "Start Rule does not match start of test string"
    with pytest.raises(StopRulesetLexing):
        actions.handle_jinja(
            jinja_analyzer, source_string, match, start_name, end_name, token_type
        )
    assert len(source_string) == jinja_analyzer.pos
    assert len(jinja_analyzer.node_buffer) == 1
    assert jinja_analyzer.node_buffer[0].token.type == token_type
    assert len(str(jinja_analyzer.node_buffer[0]).strip()) == len(source_string)


@pytest.mark.parametrize(
    "source_string",
    [
        "{% set my_var %}\n!\n{% endset %}",
        "{% set ns.my_var %}\n!\n{% endset %}",
    ],
)
def test_handle_jinja_set_block(default_analyzer: Analyzer, source_string: str) -> None:
    query = default_analyzer.parse_query(source_string=source_string)
    assert len(query.lines) == 3
    assert query.lines[0].nodes[0].token.type == TokenType.JINJA_BLOCK_START
    assert query.lines[1].nodes[0].token.type == TokenType.DATA
    assert query.lines[2].nodes[0].token.type == TokenType.JINJA_BLOCK_END


def test_handle_jinja_set_block_nested(default_analyzer: Analyzer) -> None:
    source_string = """
    {% set foo %}
    !
    something_else
    {% set bar %}bar{% endset %}
    {{ bar }}
    {% set baz %}
    baz
    {% endset %}
    {{ baz ~ bar }}
    {% endset %}
    """.strip()
    q = default_analyzer.parse_query(source_string=source_string)
    assert q.lines[0].nodes[0].token.type == TokenType.JINJA_BLOCK_START
    assert q.lines[1].nodes[0].token.type == TokenType.DATA
    assert q.lines[2].nodes[0].token.type == TokenType.JINJA_BLOCK_START
    assert q.lines[2].nodes[1].token.type == TokenType.DATA
    assert len(q.lines[2].nodes[1].open_jinja_blocks) == 2
    assert q.lines[2].nodes[2].token.type == TokenType.JINJA_BLOCK_END
    assert q.lines[3].nodes[0].token.type == TokenType.DATA
    assert q.lines[4].nodes[0].token.type == TokenType.JINJA_BLOCK_START
    assert q.lines[5].nodes[0].token.type == TokenType.DATA
    assert q.lines[6].nodes[0].token.type == TokenType.JINJA_BLOCK_END
    assert q.lines[7].nodes[0].token.type == TokenType.DATA
    assert q.lines[8].nodes[0].token.type == TokenType.JINJA_BLOCK_END


def test_handle_jinja_if_block(default_analyzer: Analyzer) -> None:
    source_string = """
    {% if foo == bar %}
        column_a,
    {%- elif foo < baz -%}
        column_b,
    {% endif %}
    """.strip()
    query = default_analyzer.parse_query(source_string=source_string)
    assert len(query.lines) == 5
    assert query.lines[0].nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert query.lines[2].nodes[0].token.type is TokenType.JINJA_BLOCK_KEYWORD
    assert query.lines[-1].nodes[0].token.type is TokenType.JINJA_BLOCK_END


def test_handle_jinja_if_block_nested(default_analyzer: Analyzer) -> None:
    source_string = """
    {% if foo == bar %}
        {%- if baz == qux %}
            column_a,
        {% else %}
            column_b,
        {% endif %}
    {%- else -%}
        column_c
    {% endif -%}
    """.strip()
    query = default_analyzer.parse_query(source_string=source_string)
    assert len(query.lines) == 9
    assert query.lines[0].nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert query.lines[1].nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert query.lines[3].nodes[0].token.type is TokenType.JINJA_BLOCK_KEYWORD
    assert query.lines[3].nodes[0].previous_node == query.lines[0].nodes[-1]
    assert query.lines[5].nodes[0].token.type is TokenType.JINJA_BLOCK_END
    assert query.lines[6].nodes[0].token.type is TokenType.JINJA_BLOCK_KEYWORD
    assert query.lines[6].nodes[0].previous_node is None
    assert query.lines[-1].nodes[0].token.type is TokenType.JINJA_BLOCK_END


def test_handle_jinja_for_block(default_analyzer: Analyzer) -> None:
    source_string = """
    {% for source in var('marketing_warehouse_ad_group_sources') %}
        {% set relation_source = 'stg_' + source + '_ad_groups' %}

        select
            '{{source}}' as source,
            *
            from {{ ref(relation_source) }}

            {% if not loop.last %}union all{% endif %}
    {% endfor %}
    """.strip()
    query = default_analyzer.parse_query(source_string=source_string)
    assert len(query.lines) == 10
    start_node = query.nodes[0]
    assert start_node.token.type is TokenType.JINJA_BLOCK_START
    assert query.nodes[1].open_jinja_blocks == [start_node]
    assert query.nodes[-2].token.type is TokenType.JINJA_BLOCK_END


@pytest.mark.parametrize(
    "source_string",
    [
        "{% endfor %}",
        "{% if foo %}{% endfor %}",
        "{% for foo in bar %}{% endif %}",
        "{% else %}",
    ],
)
def test_handle_jinja_end_block_raises(
    default_analyzer: Analyzer, source_string: str
) -> None:
    with pytest.raises(SqlfmtBracketError):
        _ = default_analyzer.parse_query(source_string=source_string)


@pytest.mark.parametrize(
    "source_string",
    [
        "{% for foo in bar %} {% endfor %}",
        "{% if foo %}{% endif %}",
        "{% if foo %} {% else %}{% endif %}",
        "{% set foo %}{% endset %}",
        "{% call noop_statement('main', status_string) %}{% endcall %}",
    ],
)
def test_handle_jinja_empty_blocks(
    default_analyzer: Analyzer, source_string: str
) -> None:
    q = default_analyzer.parse_query(source_string=source_string)
    assert len(q.lines) == 1
    assert q.nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert q.nodes[-2].token.type is TokenType.JINJA_BLOCK_END


def test_handle_jinja_call_statement_block(default_analyzer: Analyzer) -> None:
    source_string = """
    select 1,
    {% call statement() %}
    select 2 from foo
    {% endcall %}
    2
    """.strip()
    query = default_analyzer.parse_query(source_string=source_string.lstrip())

    assert query.lines[1].nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert query.lines[-2].nodes[0].token.type is TokenType.JINJA_BLOCK_END

    # ensure endcall block resets sql depth
    outer_select = query.nodes[0]
    assert query.nodes[-1].depth == (1, 0)
    assert query.nodes[-1].open_brackets == [outer_select]


def test_handle_jinja_call_block(default_analyzer: Analyzer) -> None:
    source_string = """
    select 1,
    {% call dbt_unit_testing.mock_ref('foo') %}
    a | b
    foo | bar
    {% endcall %}
    2
    """.strip()
    query = default_analyzer.parse_query(source_string=source_string.lstrip())

    assert query.lines[1].nodes[0].token.type is TokenType.JINJA_BLOCK_START
    assert query.lines[1].nodes[1].token.type is TokenType.NEWLINE
    assert query.lines[2].nodes[0].token.type is TokenType.DATA
    assert query.lines[3].nodes[0].token.type is TokenType.JINJA_BLOCK_END
    assert query.lines[4].nodes[0].token.type is TokenType.NUMBER

    # ensure call block does not change sql depth
    outer_select = query.nodes[0]
    assert query.nodes[-1].depth == (1, 0)
    assert query.nodes[-1].open_brackets == [outer_select]


def test_handle_unsupported_ddl(default_analyzer: Analyzer) -> None:
    source_string = """
    create table foo (bar int);
    select create, insert from baz;
    create table bar (foo int);
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    assert len(query.lines) == 3
    first_create_line = query.lines[0]
    assert len(first_create_line.nodes) == 3  # data, semicolon, newline
    assert first_create_line.nodes[0].token.type is TokenType.DATA
    assert first_create_line.nodes[-2].token.type is TokenType.SEMICOLON

    select_line = query.lines[1]
    assert len(select_line.nodes) == 8
    assert select_line.nodes[1].token.type is TokenType.NAME
    assert select_line.nodes[3].token.type is TokenType.NAME


def test_handle_explain(default_analyzer: Analyzer) -> None:
    source_string = """
    explain select 1;
    select explain, 1 from baz;
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    assert len(query.lines) == 2
    explain_line = query.lines[0]
    assert len(explain_line.nodes) == 5
    assert explain_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD
    assert explain_line.nodes[1].token.type is TokenType.UNTERM_KEYWORD

    select_line = query.lines[1]
    assert len(select_line.nodes) == 8
    assert select_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD
    assert select_line.nodes[1].token.type is TokenType.NAME


def test_handle_semicolon(default_analyzer: Analyzer) -> None:
    source_string = """
    select 1;
    create function foo() as select 1;
    select 1;
    create function foo() as $$ select 1 $$;
    select 1;
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    assert len(query.lines) == 5
    for line in query.lines:
        assert line.nodes[0].is_unterm_keyword


def test_handle_ddl_as_unquoted(function_analyzer: Analyzer) -> None:
    source_string = """
    create function foo
    language sql
    as select
    """
    function_analyzer.lex(source_string=source_string.lstrip())
    assert len(function_analyzer.node_buffer) == 2
    assert function_analyzer.node_buffer[0].is_unterm_keyword
    # ensure we're lexing using the main ruleset (implying an empty rule stack)
    assert function_analyzer.node_buffer[1].is_unterm_keyword
    assert not function_analyzer.rule_stack


def test_handle_ddl_as_quoted(function_analyzer: Analyzer) -> None:
    source_string = """
    create function foo
    language sql
    as $$select
    1$$ security definer
    """
    function_analyzer.lex(source_string=source_string.lstrip())
    assert len(function_analyzer.node_buffer) == 3
    assert function_analyzer.node_buffer[0].is_unterm_keyword
    # ensure we're lexing using the create_fn rules (implying an empty rule stack)
    assert function_analyzer.node_buffer[1].token.type is TokenType.QUOTED_NAME
    assert function_analyzer.rule_stack
    # ensure security definer is being lexed with the create_function ruleset (as a kw)
    assert function_analyzer.node_buffer[2].is_unterm_keyword


def test_handle_closing_angle_bracket(default_analyzer: Analyzer) -> None:
    source_string = """
    table<a int, b int>,
    array<struct<a int, b string>>,
    x >> 2,
    yr > 2022,
    foo >>= 'bar',
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    assert all([line.depth == (0, 0) for line in query.lines])
    table_line = query.lines[0]
    assert table_line.nodes[0].is_opening_bracket
    assert table_line.nodes[-3].is_closing_bracket
    array_line = query.lines[1]
    assert array_line.nodes[0].is_opening_bracket
    assert array_line.nodes[-3].is_closing_bracket
    assert array_line.nodes[-4].is_closing_bracket
    assert all([line.nodes[1].is_operator for line in query.lines[2:]])


def test_handle_number_unary(default_analyzer: Analyzer) -> None:
    source_string = """
    select
        +1,
        -2,
        -1 + -2,
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    numbers = [str(n).strip() for n in query.nodes if n.token.type is TokenType.NUMBER]
    assert numbers == ["+1", "-2", "-1", "-2"]


def test_handle_number_binary(default_analyzer: Analyzer) -> None:
    source_string = """
    select
        1 +1,
        1 -1,
        -1+2,
        something-2,
        (something)+2,
        case when true then foo else bar end+2
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    numbers = [str(n).strip() for n in query.nodes if n.token.type is TokenType.NUMBER]
    assert numbers == ["1", "1", "1", "1", "-1", "2", "2", "2", "2"]


def test_handle_nested_dictionary_in_jinja_expression(
    jinja_analyzer: Analyzer,
) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_actions/test_handle_potentially_nested_tokens.sql"
    )
    match = re.match(r"(\{\{)", source_string)
    assert match
    actions.handle_potentially_nested_tokens(
        analyzer=jinja_analyzer,
        source_string=source_string,
        match=match,
        start_name="jinja_expression_start",
        end_name="jinja_expression_end",
        token_type=TokenType.JINJA_EXPRESSION,
    )
    assert jinja_analyzer.pos == 355


def test_handle_reserved_keywords(default_analyzer: Analyzer) -> None:
    source_string = """
    select case;
    select foo.case;
    select foo.select;
    interval;
    foo.interval;
    explain;
    foo.explain;
    """
    query = default_analyzer.parse_query(source_string=source_string.lstrip())
    assert len(query.lines) == 7
    case_line = query.lines[0]
    assert len(case_line.nodes) == 4
    assert case_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD
    assert case_line.nodes[1].token.type is TokenType.STATEMENT_START

    case_name_line = query.lines[1]
    assert len(case_name_line.nodes) == 6
    assert case_name_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD
    assert case_name_line.nodes[1].token.type is TokenType.NAME
    assert case_name_line.nodes[2].token.type is TokenType.DOT
    assert case_name_line.nodes[3].token.type is TokenType.NAME

    select_name_line = query.lines[2]
    assert len(select_name_line.nodes) == 6
    assert select_name_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD
    assert select_name_line.nodes[1].token.type is TokenType.NAME
    assert select_name_line.nodes[2].token.type is TokenType.DOT
    assert select_name_line.nodes[3].token.type is TokenType.NAME

    interval_line = query.lines[3]
    assert len(interval_line.nodes) == 3
    assert interval_line.nodes[0].token.type is TokenType.WORD_OPERATOR

    interval_name_line = query.lines[4]
    assert len(interval_name_line.nodes) == 5
    assert interval_name_line.nodes[0].token.type is TokenType.NAME
    assert interval_name_line.nodes[1].token.type is TokenType.DOT
    assert interval_name_line.nodes[2].token.type is TokenType.NAME

    explain_line = query.lines[5]
    assert len(explain_line.nodes) == 3
    assert explain_line.nodes[0].token.type is TokenType.UNTERM_KEYWORD

    explain_name_line = query.lines[6]
    assert len(explain_name_line.nodes) == 5
    assert explain_name_line.nodes[0].token.type is TokenType.NAME
    assert explain_name_line.nodes[1].token.type is TokenType.DOT
    assert explain_name_line.nodes[2].token.type is TokenType.NAME



import locale
import re
import subprocess
import sys
from pathlib import Path
from typing import List

import pytest
from click.testing import CliRunner

from sqlfmt.cli import sqlfmt as sqlfmt_main
from tests.util import copy_config_file_to_dst


def run_cli_command(commands: List[str]) -> subprocess.CompletedProcess:
    process = subprocess.run(
        commands,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
        shell=True,
    )
    return process


@pytest.mark.skipif(
    sys.platform.startswith("win"),
    reason="Fails on GHA windows runner, can't repro locally",
)
@pytest.mark.parametrize(
    "cmd",
    [
        "sqlfmt --no-progressbar",
        "python -m sqlfmt --no-progressbar",
    ],
)
def test_click_cli_runner_is_equivalent_to_py_subprocess(
    sqlfmt_runner: CliRunner, cmd: str
) -> None:
    builtin_results = run_cli_command([cmd])
    click_results = sqlfmt_runner.invoke(sqlfmt_main)

    assert builtin_results.returncode == click_results.exit_code
    assert builtin_results.stdout == click_results.stdout
    assert builtin_results.stderr == click_results.stderr

    assert "https://sqlfmt.com" in click_results.stderr
    assert "sqlfmt ." in click_results.stderr


def test_help_command(sqlfmt_runner: CliRunner) -> None:
    # Sally installs sqlfmt; not knowing where to start, she types "sqlfmt --help" into
    # her command line, and sees that it displays the version and a help menu
    help_option = "--help"
    help_results = sqlfmt_runner.invoke(sqlfmt_main, args=help_option)
    assert help_results.exit_code == 0
    assert help_results.stdout.startswith("Usage: sqlfmt")


def test_version_command(sqlfmt_runner: CliRunner) -> None:
    version_option = "--version"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=version_option)
    assert results.exit_code == 0
    assert results.stdout.startswith("sqlfmt, version ")

    semver_pattern = r"\d+\.\d+.\d+"
    match = re.search(semver_pattern, results.stdout)
    assert match, "Semantic version number not in output"


def test_stdin(sqlfmt_runner: CliRunner) -> None:
    stream_input = "select 1"
    results = sqlfmt_runner.invoke(sqlfmt_main, args="-", input=stream_input)
    assert results.exit_code == 0
    assert results.stdout == "select 1\n"


def test_preformatted_check(sqlfmt_runner: CliRunner, preformatted_dir: Path) -> None:
    args = f"{preformatted_dir.as_posix()} --check"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0
    assert "passed formatting check" in results.stderr

    args = f"{preformatted_dir.as_posix()}"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args, env={"SQLFMT_CHECK": "1"})
    assert results.exit_code == 0
    assert "passed formatting check" in results.stderr


def test_preformatted_short_lines_env(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()}"
    results = sqlfmt_runner.invoke(
        sqlfmt_main, args=args, env={"SQLFMT_LINE_LENGTH": "1"}
    )
    assert results.exit_code == 0
    print(results.stderr)
    assert "5 files formatted" in results.stderr

    # test that CLI flag overrides ENV VAR
    args = f"{preformatted_dir.as_posix()} -l 88 --check"
    results = sqlfmt_runner.invoke(
        sqlfmt_main, args=args, env={"SQLFMT_LINE_LENGTH": "1"}
    )
    assert results.exit_code == 0
    print(results.stderr)
    assert "6 files passed formatting check" in results.stderr


def test_unformatted_check(sqlfmt_runner: CliRunner, unformatted_dir: Path) -> None:
    args = f"{unformatted_dir.as_posix()} --check"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 1
    assert "failed formatting check" in results.stderr

    args = f"{unformatted_dir.as_posix()}"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args, env={"SQLFMT_CHECK": "1"})
    assert results.exit_code == 1
    assert "failed formatting check" in results.stderr


def test_error_check(sqlfmt_runner: CliRunner, error_dir: Path) -> None:
    args = f"{error_dir.as_posix()} --check"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 2


def test_preformatted_single_process(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()} --single-process"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


def test_preformatted_config_file(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    # config file sets line length to 100 and enables check mode
    copy_config_file_to_dst("valid_sqlfmt_config.toml", preformatted_dir)
    args = f"{preformatted_dir.as_posix()}"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    # 3 files should fail formatting with longer line length in config
    assert results.exit_code == 1
    assert results.stderr.startswith("3 files failed formatting check")
    # supply CLI args to override config file so checks pass
    args = f"{preformatted_dir.as_posix()} --line-length 88"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


def test_preformatted_exclude_all(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = (
        f"{preformatted_dir.as_posix()} --exclude {preformatted_dir.as_posix()}/*.sql"
    )
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0
    assert results.stderr.startswith("0 files left unchanged")


def test_preformatted_clickhouse(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()} --check --dialect clickhouse"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


def test_preformatted_no_progressbar(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()} --check --no-progressbar"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


@pytest.mark.parametrize("option", ["--fast", "--safe"])
def test_preformatted_fast_safe(
    sqlfmt_runner: CliRunner, preformatted_dir: Path, option: str
) -> None:
    args = f"{preformatted_dir.as_posix()} --check {option}"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


def test_preformatted_utf_8_sig_encoding(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()} --check --encoding utf-8-sig"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0


def test_preformatted_inherit_encoding(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    args = f"{preformatted_dir.as_posix()} --check --encoding inherit"
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    if locale.getpreferredencoding().lower().replace("-", "_") == "utf_8":
        assert results.exit_code == 0
    else:
        # this directory includes a file that starts with a BOM. We'll
        # get a weird symbol if decoded with anything other than utf-8,
        # like cp-1252, which is the default on many Windows machines
        assert results.exit_code == 2
        assert results.stderr.startswith("1 file had errors")
        assert "006_has_bom.sql" in results.stderr
        assert "Could not parse SQL at position 1" in results.stderr


def test_config_option(sqlfmt_runner: CliRunner, preformatted_dir: Path) -> None:
    copy_config_file_to_dst("valid_sqlfmt_config.toml", preformatted_dir)
    args = (
        f"{preformatted_dir.as_posix()} "
        f"--config {(preformatted_dir / 'pyproject.toml').as_posix()} "
        "--check"
    )
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    # 3 files should fail formatting with longer line length in config
    assert results.exit_code == 1
    assert results.stderr.startswith("3 files failed formatting check")

    args = f"{preformatted_dir.as_posix()} --check"
    results = sqlfmt_runner.invoke(
        sqlfmt_main,
        args=args,
        env={"SQLFMT_CONFIG": f"{preformatted_dir.as_posix()}/pyproject.toml"},
    )
    assert results.exit_code == 1
    assert results.stderr.startswith("3 files failed formatting check")

    # supply CLI args to override config file so checks pass
    args = (
        f"{preformatted_dir.as_posix()} "
        f"--config {(preformatted_dir / 'pyproject.toml').as_posix()} "
        "--line-length 88 --check"
    )
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 0

    # supply CLI args to override config file so checks pass
    args = f"{preformatted_dir.as_posix()} --line-length 88 --check"
    results = sqlfmt_runner.invoke(
        sqlfmt_main,
        args=args,
        env={"SQLFMT_CONFIG": f"{preformatted_dir.as_posix()}/pyproject.toml"},
    )
    assert results.exit_code == 0


def test_config_does_not_exist(
    sqlfmt_runner: CliRunner, preformatted_dir: Path
) -> None:
    # make sure sqlfmt fails fast if the passed config doesn't exist
    args = (
        f"--config {preformatted_dir.as_posix()}/does_not_exist.toml "
        f"{preformatted_dir.as_posix()}"
    )
    results = sqlfmt_runner.invoke(sqlfmt_main, args=args)
    assert results.exit_code == 2
    assert "Error: Invalid value for '--config'" in results.stderr
    assert "does not exist" in results.stderr

    args = f"{preformatted_dir.as_posix()}"
    results = sqlfmt_runner.invoke(
        sqlfmt_main,
        args=args,
        env={"SQLFMT_CONFIG": f"{preformatted_dir.as_posix()}/does_not_exist.toml"},
    )
    assert results.exit_code == 2
    assert "Error: Invalid value for '--config'" in results.stderr
    assert "does not exist" in results.stderr



import difflib
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional

import click

from sqlfmt.exception import SqlfmtError
from sqlfmt.mode import Mode

STDIN_PATH = Path("-")


def style_output(
    msg: str, fg: Optional[str] = None, bg: Optional[str] = None, bold: bool = False
) -> str:
    """
    A thin wrapper around click.style.

    See https://click.palletsprojects.com/en/8.0.x/api/?highlight=style#click.style
    """
    s: str = click.style(msg, fg=fg, bg=bg, bold=bold)
    return s


def unstyle_output(msg: str) -> str:
    """
    A thin wrapper around click.unstyle.
    """
    s: str = click.unstyle(msg)
    return s


def display_output(msg: str, err: bool = True, nl: bool = True) -> None:
    """
    A thin wrapper around click.echo; defaults to stderr with trailing newline.
    """
    click.echo(msg, err=err, nl=nl)


@dataclass
class SqlFormatResult:
    """
    A SqlfmtResult is a summary of the changes made by sqlfmt to a single file.
    """

    source_path: Path
    source_string: str
    formatted_string: str
    encoding: str
    utf_bom: str
    exception: Optional[SqlfmtError] = None
    from_cache: bool = False

    def __post_init__(self) -> None:
        try:
            self.display_path = self.source_path.relative_to(Path.cwd())
        except ValueError:
            self.display_path = self.source_path

    def maybe_print_to_stdout(self) -> None:
        """
        If sqlfmt received a query via stdin, print the formatted string to stdout
        without an additional trailing newline.
        """
        if self.source_path == STDIN_PATH:
            display_output(self.formatted_string, err=False, nl=False)

    @property
    def has_changed(self) -> bool:
        return self.source_string != self.formatted_string

    @property
    def has_error(self) -> bool:
        return self.exception is not None


@dataclass
class Report:
    """
    An abstraction for a summary of results generated by a sqlfmt run.
    Can be printed to stderr using display_report()
    """

    results: List[SqlFormatResult]
    mode: Mode

    def __str__(self) -> str:
        """
        Returns the full contents of the Report
        """

        report = []
        if self.mode.check:
            formatted = "failed formatting check"
            unchanged = "passed formatting check"
        elif self.mode.diff:
            formatted = "would be formatted"
            unchanged = "would be left unchanged"
        else:
            formatted = "formatted"
            unchanged = "left unchanged"

        # Calculate and display summary statistics
        if not self.mode.quiet:
            if self.number_errored > 0:
                error_msg = (
                    f"{self._pluralize_file(self.number_errored)} had errors while "
                    f"formatting."
                )
                report.append(style_output(error_msg, fg="red", bold=True))
            if self.number_changed > 0:
                changed_msg = (
                    f"{self._pluralize_file(self.number_changed)} {formatted}."
                )
                report.append(style_output(changed_msg, bold=True))
            report.append(f"{self._pluralize_file(self.number_unchanged)} {unchanged}.")

        # If configured, display detailed changes
        max_errors = 1 if self.mode.quiet else 50
        for res in self.errored_results[0:max_errors]:
            err = style_output(str(res.exception), fg="red")
            report.append(f"{res.display_path}\n    {err}")

        if not self.mode.quiet:
            for res in self.changed_results:
                report.append(f"{res.display_path} {formatted}.")

        if self.mode.verbose:
            for res in self.unchanged_results:
                report.append(f"{res.display_path} {unchanged}.")

        if self.mode.diff:
            for res in self.changed_results:
                report.append(self._generate_diff(res))

        msg = "\n".join(report)
        if self.mode.color is False:
            msg = unstyle_output(msg)

        return msg

    @staticmethod
    def _pluralize_file(n: int) -> str:
        """
        Returns either "1 file" or "n files", depending on n
        """
        suffix = "s" if n != 1 else ""
        return f"{n} file{suffix}"

    @classmethod
    def _generate_diff(cls, result: SqlFormatResult) -> str:
        """
        Returns a non-colorized diff of the source and formatted
        strings in the SqlfmtResult
        """
        cleaned_lines = []
        # Work around https://bugs.python.org/issue2142
        for line in difflib.unified_diff(
            result.source_string.splitlines(keepends=True),
            result.formatted_string.splitlines(keepends=True),
            fromfile="STDIN"
            if result.source_path == STDIN_PATH
            else str(result.display_path),
            tofile="STDOUT"
            if result.source_path == STDIN_PATH
            else str(result.display_path),
            fromfiledate="(Source Query)",
            tofiledate="(Formatted Query)",
        ):
            if line[-1] == "\n":
                cleaned_lines.append(cls._style_diff_line(line))
            else:
                cleaned_lines.append(cls._style_diff_line(line + "\n"))
                cleaned_lines.append(
                    cls._style_diff_line("\\ No newline at end of file\n")
                )

        return "".join(cleaned_lines)

    @staticmethod
    def _style_diff_line(line: str) -> str:
        """
        Colorizes the diff created by _generate_diff
        """
        if line.startswith("@@"):
            styled = style_output(line, fg="cyan")
        elif line.startswith("+"):
            styled = style_output(line, fg="green")
        elif line.startswith("-"):
            styled = style_output(line, fg="red")
        else:
            styled = line
        return styled

    def display_report(self) -> None:
        """
        If sqlfmt received a query via stdin, print the formatted string to stdout.

        Then print the report that summarizes all results
        """
        if not self.mode.check and not self.mode.diff:
            for res in self.results:
                res.maybe_print_to_stdout()

        report = str(self)
        # The report can be empty (e.g., quiet mode). In
        # that case, don't print a blank line to stderr.
        if report:
            display_output(report, err=True)

    @property
    def changed_results(self) -> List[SqlFormatResult]:
        return self._filtered_results(has_changed=True, has_error=False)

    @property
    def unchanged_results(self) -> List[SqlFormatResult]:
        return self._filtered_results(has_changed=False, has_error=False)

    @property
    def errored_results(self) -> List[SqlFormatResult]:
        return self._filtered_results(has_changed=True, has_error=True)

    def _filtered_results(
        self, has_changed: bool = True, has_error: bool = False
    ) -> List[SqlFormatResult]:
        filtered = [
            r
            for r in self.results
            if r.has_changed == has_changed and r.has_error == has_error
        ]
        return sorted(filtered, key=lambda res: res.source_path)

    @property
    def number_changed(self) -> int:
        return len(self.changed_results)

    @property
    def number_unchanged(self) -> int:
        return len(self.unchanged_results)

    @property
    def number_errored(self) -> int:
        return len(self.errored_results)



import pytest

from sqlfmt.exception import SqlfmtBracketError
from sqlfmt.mode import Mode
from sqlfmt.node_manager import NodeManager
from sqlfmt.tokens import Token, TokenType
from tests.util import read_test_data


@pytest.fixture
def node_manager(default_mode: Mode) -> NodeManager:
    return NodeManager(default_mode.dialect.case_sensitive_names)


def test_calculate_depth(node_manager: NodeManager) -> None:
    select_t = Token(
        type=TokenType.UNTERM_KEYWORD,
        prefix="",
        token="select",
        spos=0,
        epos=6,
    )
    select_n = node_manager.create_node(token=select_t, previous_node=None)
    assert (select_n.depth, select_n.open_brackets) == ((0, 0), [])

    open_paren_t = Token(
        type=TokenType.BRACKET_OPEN,
        prefix="    ",
        token="(",
        spos=8,
        epos=9,
    )
    open_paren_n = node_manager.create_node(token=open_paren_t, previous_node=select_n)
    assert (open_paren_n.depth, open_paren_n.open_brackets) == ((1, 0), [select_n])

    one_t = Token(
        type=TokenType.NUMBER,
        prefix=" ",
        token="1",
        spos=10,
        epos=11,
    )
    one_n = node_manager.create_node(token=one_t, previous_node=open_paren_n)
    assert (one_n.depth, one_n.open_brackets) == ((2, 0), [select_n, open_paren_n])

    close_paren_t = Token(
        type=TokenType.BRACKET_CLOSE,
        prefix="",
        token=")",
        spos=11,
        epos=12,
    )
    close_paren_n = node_manager.create_node(token=close_paren_t, previous_node=one_n)
    assert (close_paren_n.depth, close_paren_n.open_brackets) == ((1, 0), [select_n])


def test_calculate_depth_exception(node_manager: NodeManager) -> None:
    close_paren = Token(
        type=TokenType.BRACKET_CLOSE,
        prefix="",
        token=")",
        spos=0,
        epos=1,
    )
    with pytest.raises(SqlfmtBracketError):
        _ = node_manager.create_node(close_paren, previous_node=None)


def test_jinja_depth(default_mode: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_node_manager/test_jinja_depth.sql"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    expected = [
        (0, 0),  # {{ config(materialized="table") }}
        (0, 0),  # \n
        (0, 0),  # \n
        (0, 0),  # {%- set n = 5 -%}
        (0, 0),  # \n
        (0, 0),  # with
        (1, 0),  # \n
        (1, 0),  # {% for i in range(n) %}
        (1, 1),  # \n
        (1, 1),  # dont_do_this_
        (1, 1),  # {{ i }}
        (1, 1),  # as
        (1, 1),  # (
        (2, 1),  # \n
        (2, 1),  # {% if foo %}
        (2, 2),  # \n
        (2, 2),  # select
        (3, 2),  # \n
        (2, 1),  # {% elif bar %}
        (2, 2),  # \n
        (2, 2),  # select distinct
        (3, 2),  # \n
        (2, 1),  # {% elif baz %}
        (2, 2),  # \n
        (2, 2),  # select top 25
        (3, 2),  # \n
        (2, 1),  # {% else %}
        (2, 2),  # \n
        (2, 2),  # select
        (3, 2),  # \n
        (3, 1),  # {% endif %}
        (3, 1),  # \n
        (3, 1),  # my_col
        (3, 1),  # \n
        (2, 1),  # from
        (3, 1),  # \n
        (3, 1),  # {% if i == qux %}
        (3, 2),  # \n
        (3, 2),  # zip
        (3, 2),  # \n
        (3, 1),  # {% else %}
        (3, 2),  # \n
        (3, 2),  # zap
        (3, 2),  # \n
        (3, 1),  # {% endif %}
        (3, 1),  # \n
        (1, 1),  # )
        (1, 1),  # {% if not loop.last %}
        (1, 2),  # ,
        (1, 1),  # {% endif%}
        (1, 1),  # \n
        (1, 0),  # {% endfor %}
        (1, 0),  # \n
        (1, 0),  # {% for i in range(n) %}
        (1, 1),  # \n
        (0, 1),  # select
        (1, 1),  # \n
        (1, 1),  # *
        (1, 1),  # \n
        (0, 1),  # from
        (1, 1),  # \n
        (1, 1),  # dont_do_this_
        (1, 1),  # {{ i }}
        (1, 1),  # \n
        (1, 1),  # {% if not loop.last -%}
        (1, 2),  # \n
        (0, 2),  # union all
        (0, 2),  # \n
        (0, 1),  # {%- endif %}
        (0, 1),  # \n
        (0, 0),  # {% endfor %}
        (0, 0),  # \n
    ]
    actual = [node.depth for node in q.nodes]
    assert actual == expected


def test_create_node_raises_bracket_error_on_jinja_block_end(
    node_manager: NodeManager,
) -> None:
    t = Token(
        type=TokenType.JINJA_BLOCK_END,
        prefix="",
        token="{% endif %}",
        spos=0,
        epos=11,
    )
    with pytest.raises(SqlfmtBracketError):
        _ = node_manager.create_node(t, previous_node=None)


def test_union_depth(default_mode: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_node_manager/test_union_depth.sql"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)

    expected = [
        (0, 0),  # select,
        (1, 0),  # 1,
        (1, 0),  # \n,
        (1, 0),  # \n,
        (0, 0),  # union,
        (0, 0),  # \n,
        (0, 0),  # \n,
        (0, 0),  # select,
        (1, 0),  # 2,
        (1, 0),  # \n,
        (1, 0),  # \n,
        (0, 0),  # union all,
        (0, 0),  # \n,
        (0, 0),  # \n,
        (0, 0),  # (,
        (1, 0),  # \n,
        (1, 0),  # select,
        (2, 0),  # 3,
        (2, 0),  # \n,
        (0, 0),  # ),
        (0, 0),  # \n'
    ]
    actual_depth = [n.depth for n in q.nodes]
    assert actual_depth == expected


def test_capitalization(default_mode: Mode) -> None:
    source_string = (
        'SELECT A, B, "C", {{ D }}, e, \'f\', \'G\'\nfROM "H"."j" Join I ON k And L\n'
    )
    expected = (
        'select a, b, "C", {{ D }}, e, \'f\', \'G\'\nfrom "H"."j" join i on k and l\n'
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert parsed_string == expected


@pytest.mark.parametrize(
    "source_string,expected",
    [
        (
            "SELECT A, B, \"C\", {{ D }}, e, 'f', 'G'\n"
            'fROM "H"."j" Join I ON k And L\n',
            "select A, B, \"C\", {{ D }}, e, 'f', 'G'\n"
            'from "H"."j" join I on k and L\n',
        ),
        (
            "SELECT toString(1) AS Test_string, toDateTime64('2022-05-25', 3) "
            "AS Test_dateTime64, ifNull(null, 'TestNull') as testIf, "
            "JSONExtractString('{\"abc\": \"hello\"}', 'abc') as testJSON\n",
            "select toString(1) as Test_string, toDateTime64('2022-05-25', 3) "
            "as Test_dateTime64, ifNull(null, 'TestNull') as testIf, "
            "JSONExtractString('{\"abc\": \"hello\"}', 'abc') as testJSON\n",
        ),
    ],
)
def test_capitalization_clickhouse(
    source_string: str, expected: str, clickhouse_mode: Mode
) -> None:
    q = clickhouse_mode.dialect.initialize_analyzer(
        line_length=clickhouse_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert parsed_string == expected


@pytest.mark.parametrize(
    "source_string",
    [
        "OVER",
        "IN",
        "AND",
        "EXCEPT",
        "REPLACE",
        "UNION",
        "SUM",
        "BETWEEN",
    ],
)
def test_capitalization_operators(default_mode: Mode, source_string: str) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert parsed_string.rstrip("\n") == source_string.lower()


@pytest.mark.parametrize(
    "source_string",
    [
        "1e4",
        "1E4",
        "-.1234567E+2BD",
    ],
)
def test_capitalization_numbers(default_mode: Mode, source_string: str) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert parsed_string.rstrip("\n") == source_string.lower()


@pytest.mark.parametrize(
    "source_string",
    [
        "my_schema.my_table\n",
        "my_table.*\n",
        '"my_table".*\n',
        "{{ my_schema }}.my_table\n",
        "my_schema.{{ my_table }}\n",
        "my_database.my_schema.my_table\n",
        'my_schema."my_table"\n',
        '"my_schema".my_table\n',
        '"my_schema"."my_table"\n',
        "my_table.$2\n",
        '"my_table".$2\n',
        "my_schema.{% if foo %}bar{% else %}baz{% endif %}\n",
        "json_field:key_name\n",
        'json:"KeyName"\n',
        "my_table.json_field:key_name\n",
        '"JSONField":"KeyName"::varchar\n',
        '"JSONField":"KeyName"::varchar\n',
        "my_array[1]\n",
        "my_array[1:2]\n",
        '"my_array"[1]\n',
        '"my_array"[1:2]\n',
    ],
)
def test_identifier_whitespace(default_mode: Mode, source_string: str) -> None:
    """
    Ensure we do not inject spaces into qualified identifier names
    """
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert source_string == parsed_string


@pytest.mark.parametrize(
    "source_string",
    [
        "(my_schema.my_table)\n",
        "count(*)\n",
        "count(*) over ()\n",
        "something in (somthing_else)\n",
        "some_array[offset(0)]\n",
        "some_array_func(args)[offset(0)]\n",
        "(())\n",
        "([])\n",
        "()[] + foo()[offset(1)]\n",
        "using (id)\n",
        "foo:['bar.baz']\n",
    ],
)
def test_bracket_whitespace(default_mode: Mode, source_string: str) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert source_string == parsed_string


@pytest.mark.parametrize(
    "source_string,expected_string",
    [
        ("union\n\nall", "union all\n"),
        ("union\n all", "union all\n"),
        ("union    all", "union all\n"),
        ("select\ntop\n10", "select top 10\n"),
        ("group    by", "group by\n"),
        ("not\nin", "not in\n"),
        ("not\n  similar  \n to", "not similar to\n"),
        ("right\n  outer  \n join", "right outer join\n"),
    ],
)
def test_internal_whitespace(
    default_mode: Mode, source_string: str, expected_string: str
) -> None:
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    parsed_string = "".join(str(line) for line in q.lines)
    assert parsed_string == expected_string


def test_other_whitespace(default_mode: Mode) -> None:
    source_string = "'hi'"
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)
    assert q.nodes[0].prefix == " "


def test_jinja_whitespace(default_mode: Mode) -> None:
    source_string = "{{ foo }}bar {{ foo }} bar{% if foo %} bar {% endif %}"
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)

    assert all([node.prefix == node.token.prefix for node in q.nodes])


def test_disabled_formatting(default_mode: Mode) -> None:
    source_string, _ = read_test_data(
        "unit_tests/test_node_manager/test_disabled_formatting.sql"
    )
    q = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    ).parse_query(source_string=source_string)

    selects = [n for n in q.nodes if n.is_unterm_keyword]
    assert not selects[0].formatting_disabled
    assert selects[1].formatting_disabled
    assert selects[2].formatting_disabled
    assert len(selects[2].formatting_disabled) == 2
    assert selects[3].formatting_disabled
    assert not selects[4].formatting_disabled

    # located by content rather than by index: a standalone comment inside a
    # fmt: off region now flushes its own Line, which shifts every index after it
    # without changing a byte of rendered output
    create_publication_line = next(
        line
        for line in q.lines
        if line.nodes and line.nodes[0].token.type is TokenType.DATA
    )
    assert create_publication_line.formatting_disabled
    assert create_publication_line.nodes
    create_token = create_publication_line.nodes[0].token
    assert create_token.type is TokenType.DATA
    assert create_token in create_publication_line.nodes[0].formatting_disabled
    assert len(create_publication_line.nodes[0].formatting_disabled) == 3
    semicolon_node = create_publication_line.nodes[-2]
    assert create_token not in semicolon_node.formatting_disabled



import pytest

from sqlfmt.api import format_string
from sqlfmt.mode import Mode
from tests.util import check_formatting, read_test_data


@pytest.mark.parametrize(
    "p",
    [
        "preformatted/001_select_1.sql",
        "preformatted/002_select_from_where.sql",
        "preformatted/003_literals.sql",
        "preformatted/004_with_select.sql",
        "preformatted/005_fmt_off.sql",
        "preformatted/006_fmt_off_447.sql",
        "preformatted/007_fmt_off_comments.sql",
        "preformatted/008_reserved_names.sql",
        "preformatted/009_empty.sql",
        "preformatted/010_comment_only.sql",
        "preformatted/011_triple_quotes.sql",
        "preformatted/301_multiline_jinjafmt.sql",
        "preformatted/302_jinjafmt_multiline_str.sql",
        "preformatted/303_jinjafmt_more_mutliline_str.sql",
        "preformatted/400_create_table.sql",
        "preformatted/401_create_row_access_policy.sql",
        "preformatted/402_alter_table.sql",
        "unformatted/100_select_case.sql",
        "unformatted/101_multiline.sql",
        "unformatted/102_lots_of_comments.sql",
        "unformatted/103_window_functions.sql",
        "unformatted/104_joins.sql",
        "unformatted/106_leading_commas.sql",
        "unformatted/107_jinja_blocks.sql",
        "unformatted/108_test_block.sql",
        "unformatted/109_lateral_flatten.sql",
        "unformatted/110_other_identifiers.sql",
        "unformatted/111_chained_boolean_between.sql",
        "unformatted/112_semicolons.sql",
        "unformatted/113_utils_group_by.sql",
        "unformatted/114_unions.sql",
        "unformatted/115_select_star_except.sql",
        "unformatted/116_chained_booleans.sql",
        "unformatted/117_whitespace_in_tokens.sql",
        "unformatted/118_within_group.sql",
        "unformatted/119_psycopg_placeholders.sql",
        "unformatted/120_array_literals.sql",
        "unformatted/121_stubborn_merge_edge_cases.sql",
        "unformatted/122_values.sql",
        "unformatted/123_spark_keywords.sql",
        "unformatted/124_bq_compound_types.sql",
        "unformatted/125_numeric_literals.sql",
        "unformatted/126_blank_lines.sql",
        "unformatted/127_more_comments.sql",
        "unformatted/128_double_slash_comments.sql",
        "unformatted/129_duckdb_joins.sql",
        "unformatted/130_athena_data_types.sql",
        "unformatted/131_assignment_statement.sql",
        "unformatted/132_spark_number_literals.sql",
        "unformatted/133_for_else.sql",
        "unformatted/134_databricks_type_hints.sql",
        "unformatted/135_star_columns.sql",
        "unformatted/136_databricks_variant.sql",
        "unformatted/200_base_model.sql",
        "unformatted/201_basic_snapshot.sql",
        "unformatted/202_unpivot_macro.sql",
        "unformatted/203_gitlab_email_domain_type.sql",
        "unformatted/204_gitlab_tag_validation.sql",
        "unformatted/205_rittman_hubspot_deals.sql",
        "unformatted/206_gitlab_prep_geozone.sql",
        "unformatted/207_rittman_int_journals.sql",
        "unformatted/208_rittman_int_plan_breakout_metrics.sql",
        "unformatted/209_rittman_int_web_events_sessionized.sql",
        "unformatted/210_gitlab_gdpr_delete.sql",
        "unformatted/211_http_2019_cdn_17_20.sql",
        "unformatted/212_http_2019_cms_14_02.sql",
        "unformatted/213_gitlab_fct_sales_funnel_target.sql",
        "unformatted/214_get_unique_attributes.sql",
        "unformatted/215_gitlab_get_backup_table_command.sql",
        "unformatted/216_gitlab_zuora_revenue_revenue_contract_line_source.sql",
        "unformatted/217_dbt_unit_testing_csv.sql",
        "unformatted/218_multiple_c_comments.sql",
        "unformatted/219_any_all_agg.sql",
        "unformatted/220_clickhouse_joins.sql",
        "unformatted/300_jinjafmt.sql",
        "unformatted/400_create_fn_and_select.sql",
        "unformatted/401_explain_select.sql",
        "unformatted/402_delete_from_using.sql",
        "unformatted/403_grant_revoke.sql",
        "unformatted/404_create_function_pg_examples.sql",
        "unformatted/405_create_function_snowflake_examples.sql",
        "unformatted/406_create_function_bq_examples.sql",
        "unformatted/407_alter_function_pg_examples.sql",
        "unformatted/408_alter_function_snowflake_examples.sql",
        "unformatted/409_create_external_function.sql",
        "unformatted/410_create_warehouse.sql",
        "unformatted/411_create_clone.sql",
        "unformatted/412_pragma.sql",
        "unformatted/900_create_view.sql",
        "unformatted/998_unsupported_ddl_with_jinja.sql",
        "unformatted/999_unsupported_ddl.sql",
    ],
)
def test_formatting(p: str) -> None:
    mode = Mode()

    source, expected = read_test_data(p)
    actual = format_string(source, mode)

    check_formatting(expected, actual, ctx=p)

    second_pass = format_string(actual, mode)
    check_formatting(expected, second_pass, ctx=f"2nd-{p}")



import codecs
import io
import os
from pathlib import Path
from typing import Any, List, Set, Type

import pytest
from tqdm import tqdm

from sqlfmt.api import (
    _format_many,
    _perform_safety_check,
    _read_path_or_stdin,
    _update_source_files,
    format_string,
    get_matching_paths,
    initialize_progress_bar,
    run,
)
from sqlfmt.exception import (
    SqlfmtBracketError,
    SqlfmtEquivalenceError,
    SqlfmtError,
    SqlfmtUnicodeError,
)
from sqlfmt.mode import Mode


@pytest.fixture
def preformatted_files(preformatted_dir: Path) -> List[Path]:
    return list(preformatted_dir.iterdir())


@pytest.fixture
def unformatted_files(unformatted_dir: Path) -> List[Path]:
    return list(unformatted_dir.iterdir())


@pytest.fixture
def file_discovery_dir() -> Path:
    return Path("tests/data/unit_tests/test_api/test_file_discovery").resolve()


@pytest.fixture
def all_files(file_discovery_dir: Path) -> Set[Path]:
    p = file_discovery_dir
    files = {
        p / "top_level_file.sql",
        p / "top_level_file.two.sql",
        p / "a_directory/one_file.sql",
        p / "a_directory/nested_directory/another_file.sql",
        p / "a_directory/nested_directory/j2_extension.sql.jinja",
        p / "a_directory/symlink_source_directory/symlink_file.sql",
        p / "a_directory/symlink_target_directory/symlink_file.sql",
    }
    return files


@pytest.fixture
def sql_jinja_files(file_discovery_dir: Path) -> Set[Path]:
    p = file_discovery_dir
    files = {p / "a_directory/nested_directory/j2_extension.sql.jinja"}
    return files


def test_file_discovery(
    default_mode: Mode, file_discovery_dir: Path, all_files: Set[Path]
) -> None:
    res = get_matching_paths(file_discovery_dir.iterdir(), default_mode)

    assert res == all_files


@pytest.mark.parametrize(
    "exclude",
    [
        ["**/*_file*"],
        ["**/*.sql"],
        [
            "**/top*",
            "**/a_directory/*",
            "**/a_directory/**/another_file.sql",
            "**/a_directory/**/symlink_file.sql",
        ],
    ],
)
def test_file_discovery_with_excludes(
    exclude: List[str], file_discovery_dir: Path, sql_jinja_files: Set[Path]
) -> None:
    mode = Mode(exclude=exclude, exclude_root=file_discovery_dir)
    res = get_matching_paths(file_discovery_dir.iterdir(), mode)
    assert res == sql_jinja_files


def test_file_discovery_with_abs_excludes(
    file_discovery_dir: Path, sql_jinja_files: Set[Path]
) -> None:
    exclude = [str(file_discovery_dir / "**/*.sql")]
    mode = Mode(exclude=exclude, exclude_root=None)
    res = get_matching_paths(file_discovery_dir.iterdir(), mode)
    assert res == sql_jinja_files


def test_file_discovery_with_invalid_excludes(
    file_discovery_dir: Path, all_files: Set[Path]
) -> None:
    exclude = ["."]
    for root in [file_discovery_dir, None]:
        mode = Mode(exclude=exclude, exclude_root=root)
        res = get_matching_paths(file_discovery_dir.iterdir(), mode)
        assert res == all_files


def test_file_discovery_with_excludes_no_root(
    file_discovery_dir: Path, all_files: Set[Path], sql_jinja_files: Set[Path]
) -> None:
    mode = Mode(exclude=["**/*.sql"], exclude_root=None)

    # relative to here, excludes shouldn't do anything.
    cwd = os.getcwd()
    try:
        os.chdir(Path(__file__).parent)
        res = get_matching_paths(file_discovery_dir.iterdir(), mode)
    finally:
        os.chdir(cwd)

    assert res == all_files

    # relative to file_discovery_dir, excludes should knock out most files
    try:
        os.chdir(file_discovery_dir)
        res = get_matching_paths(file_discovery_dir.iterdir(), mode)
    finally:
        os.chdir(cwd)

    assert res == sql_jinja_files


def test_format_empty_string(all_output_modes: Mode) -> None:
    source = expected = ""
    actual = format_string(source, all_output_modes)
    assert expected == actual


@pytest.mark.parametrize(
    "source,exception",
    [
        ("select )\n", SqlfmtBracketError),
        ("{{\n", SqlfmtBracketError),
    ],
)
def test_format_bad_string(
    all_output_modes: Mode, source: str, exception: Type[SqlfmtError]
) -> None:
    with pytest.raises(exception):
        _ = format_string(source, all_output_modes)


def test_format_many_preformatted(
    preformatted_files: List[Path], all_output_modes: Mode
) -> None:
    results = list(_format_many(preformatted_files, {}, all_output_modes))

    assert len(results) == len(preformatted_files), (
        "Did not generate one result for each input file"
    )
    assert not any([res.has_changed for res in results]), "Changed preformatted file"
    assert all(
        [
            raw == res.source_path
            for raw, res in zip(preformatted_files, results, strict=True)
        ]
    ), "Result stored a source path that doesn't match the raw path passed to api"


def test_format_many_unformatted(
    unformatted_files: List[Path], default_mode: Mode
) -> None:
    results = list(_format_many(unformatted_files, {}, default_mode))

    assert len(results) == len(unformatted_files), (
        "Did not generate one result for each input file"
    )
    assert all([res.has_changed for res in results]), "Did not change unformatted file"
    assert all(
        [
            raw == res.source_path
            for raw, res in zip(unformatted_files, results, strict=True)
        ]
    ), "Result stored a source path that doesn't match the raw path passed to api"


def test_update_source_files_preformatted(
    preformatted_files: List[Path], default_mode: Mode
) -> None:
    results = list(_format_many(preformatted_files, {}, default_mode))

    expected_last_update_timestamps = [
        os.stat(res.source_path).st_mtime for res in results if res.source_path
    ]

    assert len(results) == len(expected_last_update_timestamps)

    _update_source_files(results)

    actual_last_update_timestamps = [
        os.stat(res.source_path).st_mtime for res in results if res.source_path
    ]

    assert len(expected_last_update_timestamps) == len(actual_last_update_timestamps)

    assert all(
        [
            expected == actual
            for expected, actual in zip(
                expected_last_update_timestamps,
                actual_last_update_timestamps,
                strict=True,
            )
        ]
    ), "Should not have written a new file for an unchanged result"


def test_update_source_files_unformatted(
    unformatted_files: List[Path], default_mode: Mode
) -> None:
    results = list(_format_many(unformatted_files, {}, default_mode))

    original_update_timestamps = [
        os.stat(res.source_path).st_mtime for res in results if res.source_path
    ]

    assert len(results) == len(original_update_timestamps)

    _update_source_files(results)

    new_update_timestamps = [
        os.stat(res.source_path).st_mtime for res in results if res.source_path
    ]

    assert len(new_update_timestamps) == len(original_update_timestamps)

    assert all(
        [
            new > original
            for original, new in zip(
                original_update_timestamps, new_update_timestamps, strict=True
            )
        ]
    ), "Should have written a new file for an unchanged result"

    for res in results:
        assert res.source_path
        with open(res.source_path, "r") as f:
            new_file_contents = f.read()
        assert new_file_contents == res.formatted_string
        assert new_file_contents != res.source_string


def test_run_unformatted_update(
    unformatted_dir: Path, default_mode: Mode, monkeypatch: pytest.MonkeyPatch
) -> None:
    files = get_matching_paths([unformatted_dir], default_mode)
    # confirm that we call the _update_source function
    monkeypatch.delattr("sqlfmt.api._update_source_files")
    with pytest.raises(NameError):
        _ = run(files=files, mode=default_mode)


def test_run_preformatted(
    preformatted_files: List[Path], all_output_modes: Mode, reset_cache_mode: Mode
) -> None:
    files = preformatted_files
    report = run(files=files, mode=all_output_modes)
    assert report.number_changed == 0
    assert report.number_unchanged == len(preformatted_files)
    assert report.number_errored == 0
    assert not any([res.from_cache for res in report.results])

    cached_run = run(files=files, mode=all_output_modes)
    assert cached_run.number_changed == report.number_changed
    assert cached_run.number_unchanged == report.number_unchanged
    assert cached_run.number_errored == report.number_errored
    assert all([res.from_cache for res in cached_run.results])

    uncached_run = run(files=files, mode=reset_cache_mode)
    assert uncached_run.number_changed == report.number_changed
    assert uncached_run.number_unchanged == report.number_unchanged
    assert uncached_run.number_errored == report.number_errored
    assert not any([res.from_cache for res in uncached_run.results])


def test_run_unformatted(unformatted_files: List[Path], all_output_modes: Mode) -> None:
    files = unformatted_files
    report = run(files=files, mode=all_output_modes)
    assert report.number_changed == len(unformatted_files)
    assert report.number_unchanged == 0
    assert report.number_errored == 0
    assert not any([res.from_cache for res in report.results])

    cached_run = run(files=files, mode=all_output_modes)
    if all_output_modes.diff or all_output_modes.check:
        assert cached_run.number_changed == report.number_changed
        assert cached_run.number_unchanged == report.number_unchanged
    else:
        assert cached_run.number_changed == 0
        assert cached_run.number_unchanged == report.number_changed
    assert cached_run.number_errored == report.number_errored
    if not all_output_modes.diff and not all_output_modes.check:
        assert all([res.from_cache for res in cached_run.results])


def test_run_error(error_dir: Path, all_output_modes: Mode) -> None:
    p = [error_dir]
    files = get_matching_paths(p, all_output_modes)
    report = run(files=files, mode=all_output_modes)
    assert report.number_changed == 0
    assert report.number_unchanged == 0
    assert report.number_errored == 4
    assert not any([res.from_cache for res in report.results])

    cached_run = run(files=files, mode=all_output_modes)
    assert cached_run.number_changed == report.number_changed
    assert cached_run.number_unchanged == report.number_unchanged
    assert cached_run.number_errored == report.number_errored
    assert not any([res.from_cache for res in cached_run.results])


def test_run_stdin(all_output_modes: Mode, monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("sys.stdin", io.StringIO("select\n    1"))
    report = run(files=[Path("-")], mode=all_output_modes)
    assert report.number_changed == 1
    assert report.number_unchanged == 0
    assert report.number_errored == 0


def test_run_on_nothing(all_output_modes: Mode) -> None:
    report = run(files=[], mode=all_output_modes)
    assert report.number_changed == 0
    assert report.number_unchanged == 0
    assert report.number_errored == 0


def test_run_single_process_does_not_use_multiprocessing(
    unformatted_dir: Path, single_process_mode: Mode, monkeypatch: pytest.MonkeyPatch
) -> None:
    # confirm that we do not call _multiprocess_map; if we do,
    # this will raise
    files = get_matching_paths([unformatted_dir], single_process_mode)
    monkeypatch.delattr("sqlfmt.api._multiprocess_map")
    _ = run(files=files, mode=single_process_mode)


def test_run_with_callback(
    capsys: Any, unformatted_dir: Path, default_mode: Mode
) -> None:
    def print_dot(_: Any) -> None:
        print(".", end="", flush=True)

    files = get_matching_paths([unformatted_dir], default_mode)
    expected_dots = len(files)

    _ = run(files=files, mode=default_mode, callback=print_dot)
    captured = capsys.readouterr()

    assert "." * expected_dots in captured.out


def test_initialize_progress_bar(default_mode: Mode) -> None:
    total = 100
    progress_bar, progress_callback = initialize_progress_bar(
        total=total, mode=default_mode, force_progress_bar=True
    )
    assert progress_bar
    assert isinstance(progress_bar, tqdm)
    assert progress_bar.format_dict.get("n") == 0
    assert progress_bar.format_dict.get("total") == total

    assert progress_callback is not None
    progress_callback("foo")  # type: ignore
    assert progress_bar.format_dict.get("n") == 1


def test_initialize_disabled_progress_bar(no_progressbar_mode: Mode) -> None:
    total = 100
    progress_bar, progress_callback = initialize_progress_bar(
        total=total, mode=no_progressbar_mode, force_progress_bar=True
    )
    # a disabled progress bar's elapsed timer will not count up,
    # and calling update() will not increment n
    assert progress_bar
    assert isinstance(progress_bar, tqdm)
    assert progress_bar.format_dict.get("n") == 0
    assert progress_bar.format_dict.get("total") == total
    assert progress_bar.format_dict.get("elapsed") == 0

    assert progress_callback is not None
    progress_callback("foo")  # type: ignore
    assert progress_bar.format_dict.get("n") == 0


def test_perform_safety_check(default_mode: Mode) -> None:
    source_string = "select 1, 2, 3\n"

    analyzer = default_mode.dialect.initialize_analyzer(
        line_length=default_mode.line_length
    )
    raw_query = analyzer.parse_query(source_string)

    with pytest.raises(SqlfmtEquivalenceError) as excinfo:
        # drops last token
        _perform_safety_check(analyzer, raw_query, "select 1, 2, \n")

    assert "Raw query was 6 tokens; formatted query was 5 tokens." in str(excinfo.value)

    with pytest.raises(SqlfmtEquivalenceError) as excinfo:
        # changes a token
        _perform_safety_check(analyzer, raw_query, "select a, 2, 3\n")

    assert (
        "First mismatching token at position 1: raw: TokenType.NUMBER; "
        "result: TokenType.NAME." in str(excinfo.value)
    )

    with pytest.raises(SqlfmtEquivalenceError) as excinfo:
        # adds a comment
        _perform_safety_check(
            analyzer, raw_query, "select\n-- new comment\n    1, 2, 3\n"
        )

    assert (
        "Raw query had 0 comment characters; formatted query had 10 comment characters"
        in str(excinfo.value)
    )

    # does not raise
    _perform_safety_check(analyzer, raw_query, "select\n    1, 2, 3\n")


@pytest.mark.parametrize(
    "encoding,bom",
    [
        ("utf-8", b""),
        ("utf-8", codecs.BOM_UTF8),
        ("utf-8-sig", b""),  # encoding with utf-8-sig will add a bom
        ("utf-16", b""),
        ("utf_16_be", codecs.BOM_UTF16_BE),
        ("utf_16_le", codecs.BOM_UTF16_LE),
        ("utf-32", b""),
        ("utf_32_be", codecs.BOM_UTF32_BE),
        ("utf_32_le", codecs.BOM_UTF32_LE),
        ("cp1250", b""),
        ("cp1252", b""),
        ("latin-1", b""),
        ("ascii", b""),
    ],
)
def test_read_path_or_stdin_many_encodings(
    encoding: str, bom: bytes, tmp_path: Path
) -> None:
    p = tmp_path / "q.sql"
    # create a new file with the specified encoding and BOM
    raw_query = "select\n\n\n1\n"
    file_contents = bom + raw_query.encode(encoding)
    with open(p, "wb") as f:
        f.write(file_contents)

    mode = Mode(encoding=encoding)
    actual_source, actual_encoding, actual_bom = _read_path_or_stdin(p, mode)
    assert actual_source == raw_query
    assert actual_encoding == encoding.lower().replace("-", "_")
    assert actual_bom == bom.decode(encoding)


def test_read_path_or_stdin_error(tmp_path: Path) -> None:
    p = tmp_path / "q.sql"
    with open(p, "w", encoding="utf-8") as f:
        f.write("select 'ň' as ch")

    mode = Mode(encoding="cp1250")
    with pytest.raises(SqlfmtUnicodeError) as exc_info:
        _, _, _ = _read_path_or_stdin(p, mode)

    assert "cp1250" in str(exc_info.value)



import asyncio
import codecs
import concurrent.futures
import locale
import sys
from functools import partial
from glob import glob
from itertools import zip_longest
from pathlib import Path, PurePath
from typing import (
    Awaitable,
    Callable,
    Collection,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
)

from tqdm import tqdm

from sqlfmt.analyzer import Analyzer
from sqlfmt.cache import Cache, check_cache, clear_cache, load_cache, write_cache
from sqlfmt.exception import SqlfmtEquivalenceError, SqlfmtError, SqlfmtUnicodeError
from sqlfmt.mode import Mode as Mode
from sqlfmt.query import Query
from sqlfmt.query_formatter import QueryFormatter
from sqlfmt.report import STDIN_PATH, Report, SqlFormatResult

T = TypeVar("T")
R = TypeVar("R")


def format_string(source_string: str, mode: Mode) -> str:
    """
    Takes a raw query string and a mode as input, returns the formatted query
    as a string, or raises a SqlfmtError if the string cannot be formatted.

    If mode.fast is False, also performs a safety check to ensure no tokens
    are dropped from the original input.
    """
    analyzer = mode.dialect.initialize_analyzer(line_length=mode.line_length)
    raw_query = analyzer.parse_query(source_string=source_string)
    formatter = QueryFormatter(mode)
    formatted_query = formatter.format(raw_query)
    result = str(formatted_query)

    if not mode.fast and not mode.check and not mode.diff:
        _perform_safety_check(analyzer, raw_query, result)

    return result


def run(
    files: Collection[Path],
    mode: Mode,
    callback: Optional[Callable[[Awaitable[SqlFormatResult]], None]] = None,
) -> Report:
    """
    Runs sqlfmt on all files in Collection of Paths (files), using the specified Mode.

    Modifies sql files in place, by default. Check or diff Mode do not modify files,
    they only create a Report.

    If a callback is provided, will execute the callback after each file is formatted.

    Returns a Report that can be queried or printed.
    """

    if mode.reset_cache:
        clear_cache()
        cache = {}
    else:
        cache = load_cache()

    results = _format_many(files, cache, mode, callback=callback)

    report = Report(results, mode)

    if not (mode.check or mode.diff):
        _update_source_files(results)
    write_cache(cache, results, mode)

    return report


def get_matching_paths(paths: Iterable[Path], mode: Mode) -> Set[Path]:
    """
    Takes a list of paths (files or directories) and a mode as an input, and
    yields paths to individual files that match the input paths (or are contained in
    its directories) and are not excluded by the mode's exclude glob
    """
    include_set = _get_included_paths(paths, mode)
    exclude_set = set()

    if mode.exclude:
        for s in mode.exclude:
            if PurePath(s).is_absolute():
                exclude_set.update([Path(g) for g in glob(s, recursive=True)])
            else:
                exclude_root = mode.exclude_root or Path.cwd()
                try:
                    # Prefer glob.glob() to Path.glob() here, even though the
                    # latter would be more concise, since Path.glob() does not
                    # resolve symlinks properly in Python <3.13
                    exclude_set.update(
                        [Path(g) for g in glob(str(exclude_root / s), recursive=True)]
                    )
                except IndexError:
                    # for some reason Path.glob(".") raises an index error,
                    # although glob.glob(".") returns ["."]
                    pass

    return include_set - exclude_set


def initialize_progress_bar(
    total: int, mode: Mode, force_progress_bar: bool = False
) -> Tuple[tqdm, Callable[[Awaitable[SqlFormatResult]], None]]:
    """
    Return a callable that can be used with api.run to display a progress bar
    that updates after each file is formatted.

    Pass force_progress_bar to enable the progress bar, even on non-TTY
    terminals (this is handy for testing the progress bar).
    """
    if mode.no_progressbar:
        disable = True
    elif force_progress_bar:
        disable = False
    else:
        # will be disabled on non-TTY envs, enabled otherwise
        disable = None
    progress_bar: tqdm = tqdm(
        total=total, leave=False, disable=disable, delay=0.5, unit="file"
    )

    def progress_callback(_: Awaitable[SqlFormatResult]) -> None:
        progress_bar.update()

    return progress_bar, progress_callback


def _get_included_paths(paths: Iterable[Path], mode: Mode) -> Set[Path]:
    """
    Takes a list of absolute paths (files or directories) and a mode as an input, and
    yields paths to individual files that match the input paths (or are contained in
    its directories)
    """
    include_set = set()
    for p in paths:
        if p == STDIN_PATH:
            include_set.add(p)
        elif p.is_file() and str(p).endswith(tuple(mode.SQL_EXTENSIONS)):
            include_set.add(p)
        elif p.is_dir():
            include_set |= _get_included_paths(p.iterdir(), mode)

    return include_set


def _format_many(
    paths: Collection[Path],
    cache: Cache,
    mode: Mode,
    callback: Optional[Callable[[Awaitable[SqlFormatResult]], None]] = None,
) -> List[SqlFormatResult]:
    """
    Runs sqlfmt on all files in a collection of paths, using the specified mode.

    If there are multiple paths and the mode allows it, uses asyncio's implementation
    of multiprocessing. Otherwise, reverts to single-processing behavior

    Returns a list of SqlFormatResults. Does not write formatted Queries back to disk
    """
    results: List[SqlFormatResult] = []
    cache_misses: List[Path] = []
    for path in paths:
        cached = check_cache(cache=cache, p=path)
        if cached:
            results.append(
                SqlFormatResult(
                    source_path=path,
                    source_string="",
                    formatted_string="",
                    encoding="",
                    utf_bom="",
                    from_cache=True,
                )
            )
        else:
            cache_misses.append(path)

    format_func = partial(_format_one, mode=mode)
    if len(cache_misses) > 1 and not mode.single_process:
        results.extend(
            asyncio.run(_multiprocess_map(format_func, paths, callback=callback))
        )
    else:
        results.extend((map(format_func, cache_misses)))

    return results


async def _multiprocess_map(
    func: Callable[[T], R],
    seq: Iterable[T],
    callback: Optional[Callable[[Awaitable[R]], None]] = None,
) -> List[R]:
    """
    Using multiple processes, creates a Future for each application of func to
    an item in seq, then gathers all Futures and returns the result.

    Provides a similar interface to the map() built-in, but executes in multiple
    processes.
    """
    loop = asyncio.get_event_loop()
    with concurrent.futures.ProcessPoolExecutor() as pool:
        tasks = []
        for item in seq:
            future = loop.run_in_executor(pool, func, item)
            if callback:
                future.add_done_callback(callback)
            tasks.append(future)
        results: List[R] = await asyncio.gather(*tasks)
    return results


def _format_one(path: Path, mode: Mode) -> SqlFormatResult:
    """
    Runs format_string on the contents of a single file (found at path). Handles
    potential user errors in formatted code, and returns a SqlfmtResult
    """
    source, encoding, utf_bom = _read_path_or_stdin(path, mode)
    try:
        formatted = format_string(source, mode)
        return SqlFormatResult(
            source_path=path,
            source_string=source,
            formatted_string=formatted,
            encoding=encoding,
            utf_bom=utf_bom,
        )
    except SqlfmtError as e:
        return SqlFormatResult(
            source_path=path,
            source_string=source,
            formatted_string="",
            encoding=encoding,
            utf_bom=utf_bom,
            exception=e,
        )


def _update_source_files(results: Iterable[SqlFormatResult]) -> None:
    """
    Overwrites file contents at result.source_path with result.formatted_string.

    No-ops for unchanged files, results without a source path, and empty files
    """
    for res in results:
        if res.has_changed and res.source_path != STDIN_PATH and res.formatted_string:
            with open(res.source_path, "w", encoding=res.encoding) as f:
                f.write(res.formatted_string)


def _read_path_or_stdin(path: Path, mode: Mode) -> Tuple[str, str, str]:
    """
    If passed a Path, calls open() and read() and returns contents as
    a tuple of strings. The first element is the contents of the file; the
    second element is the encoding used to read the file; the third
    element is either the utf BOM or an empty string.

    If passed Path("-"), calls sys.stdin.read()
    """
    encoding = (
        (
            locale.getpreferredencoding()
            if mode.encoding.lower() == "inherit"
            else mode.encoding
        )
        .lower()
        .replace("-", "_")
    )
    bom_map: Dict[str, List[bytes]] = {
        "utf": [codecs.BOM_UTF8],
        "utf8": [codecs.BOM_UTF8],
        "u8": [codecs.BOM_UTF8],
        "utf16": [codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE],
        "u16": [codecs.BOM_UTF16_LE, codecs.BOM_UTF16_BE],
        "utf16le": [codecs.BOM_UTF16_LE],
        "utf16be": [codecs.BOM_UTF16_BE],
        "utf32": [codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE],
        "u32": [codecs.BOM_UTF32_LE, codecs.BOM_UTF32_BE],
        "utf32le": [codecs.BOM_UTF32_LE],
        "utf32be": [codecs.BOM_UTF32_BE],
    }
    detected_bom = ""
    if path == STDIN_PATH:
        # todo: customize encoding of stdin
        source = sys.stdin.read()
    else:
        try:
            with open(path, "r", encoding=encoding) as f:
                source = f.read()
            if encoding.startswith("utf") and encoding != "utf_8_sig":
                for b in [
                    bom.decode(encoding)
                    for bom in bom_map.get(encoding.replace("_", ""), [])
                ]:
                    if source.startswith(b):
                        detected_bom = b
                        source = source[len(b) :]
                        break

        except UnicodeDecodeError as e:
            raise SqlfmtUnicodeError(
                f"Error reading file {path}\n"
                f"File could not be decoded using {encoding}. "
                f"Specifically, {repr(e.object)} at position {e.start} failed "
                f"with: {e.reason}.\n"
                "You can specify a different encoding by running sqlfmt "
                "with the --encoding option. Or set --encoding to 'none' to "
                "use the system default encoding. We suggest always using "
                "utf-8 for all files."
            ) from e
    return source, encoding, detected_bom


def _perform_safety_check(analyzer: Analyzer, raw_query: Query, result: str) -> None:
    """
    Raises a SqlfmtEquivalenceError if re-lexing
    the result produces a different set of tokens than
    the original.
    """
    result_query = analyzer.parse_query(source_string=result)
    filtered_raw_tokens = [
        token.type for token in raw_query.tokens if token.type.is_equivalent_in_output
    ]
    filtered_result_tokens = [
        token.type
        for token in result_query.tokens
        if token.type.is_equivalent_in_output
    ]

    try:
        assert filtered_raw_tokens == filtered_result_tokens
    except AssertionError as e:
        raw_len = len(filtered_raw_tokens)
        result_len = len(filtered_result_tokens)
        mismatch_pos = 0
        mismatch_raw = ""
        mismatch_res = ""

        for i, (raw, res) in enumerate(
            zip_longest(filtered_raw_tokens, filtered_result_tokens)
        ):
            if raw is not res:
                mismatch_pos = i
                mismatch_raw = str(raw)
                mismatch_res = str(res)
                break

        raise SqlfmtEquivalenceError(
            "There was a problem formatting your query that "
            "caused the safety check to fail. Please open an "
            f"issue. Raw query was {raw_len} tokens; formatted "
            f"query was {result_len} tokens. First mismatching "
            f"token at position {mismatch_pos}: raw: {mismatch_raw}; "
            f"result: {mismatch_res}."
        ) from e

    raw_comments = [
        comment.body for line in raw_query.lines for comment in line.comments
    ]
    result_comments = [
        comment.body for line in result_query.lines for comment in line.comments
    ]
    stripped_raw = "".join(["".join(c.split()) for c in raw_comments])
    stripped_res = "".join(["".join(c.split()) for c in result_comments])
    try:
        assert stripped_raw == stripped_res
    except AssertionError as e:
        raw_len = len(stripped_raw)
        result_len = len(stripped_res)
        mismatch_pos = 0
        mismatch_raw = ""
        mismatch_res = ""

        for i, (raw, res) in enumerate(zip_longest(stripped_raw, stripped_res)):
            if raw is not res:
                mismatch_pos = i
                mismatch_raw = stripped_raw[i : i + 25]
                mismatch_res = stripped_res[i : i + 25]
                break

        raise SqlfmtEquivalenceError(
            "There was a problem formatting your query that "
            "caused the safety check to fail. Please open an "
            f"issue. Raw query had {raw_len} comment characters; formatted "
            f"query had {result_len} comment characters. First mismatching "
            f"character at position {mismatch_pos}: raw: {mismatch_raw}; "
            f"result: {mismatch_res}."
        ) from e



from sqlfmt.mode import Mode
from sqlfmt.query_formatter import QueryFormatter


def test_dedent_jinja_block_ends(default_mode: Mode) -> None:
    formatter = QueryFormatter(default_mode)
    source_string = (
        "{% if foo %}\n"
        "select\n"
        "{% else %}\n"
        "select distinct\n"
        "    {% endif %}\n"
        "    my_col\n"
    )
    raw_query = formatter.mode.dialect.initialize_analyzer(
        formatter.mode.line_length
    ).parse_query(source_string)
    depth_before = [line.depth for line in raw_query.lines]
    assert depth_before[-2] == (1, 0)
    new_lines = formatter._dedent_jinja_blocks(raw_query.lines)
    depth_after = [line.depth for line in new_lines]
    assert depth_after <= depth_before
    assert depth_after[-2] == (0, 0)


def test_dedent_jinja_blocks(default_mode: Mode) -> None:
    formatter = QueryFormatter(default_mode)
    source_string = (
        "with\n"
        "    a as (select * from a),\n"
        "    {% for i in range(n) %}\n"
        "select\n"
        "    *\n"
        "from\n"
        "    dont_do_this_{{ i }}\n"
        "    {% if not loop.last -%}\n"
        "union all\n"
        "    {%- endif %}\n"
        "    {% endfor %}\n"
    )
    raw_query = formatter.mode.dialect.initialize_analyzer(
        formatter.mode.line_length
    ).parse_query(source_string)
    new_lines = formatter._dedent_jinja_blocks(raw_query.lines)
    jinja_depths = [
        line.depth[0] for line in new_lines if line.is_standalone_jinja_statement
    ]
    assert all([line_depth == 0 for line_depth in jinja_depths])


def test_remove_extra_blank_lines(default_mode: Mode) -> None:
    formatter = QueryFormatter(default_mode)
    source_string = (
        "\n\n\n\n"
        "select 1\n;\n\n\n\n"
        "select\n    1,\n\n\n    2\n;\n"
        "{% macro foo() %}\n\n\n\n\nfoo\n{% endmacro %}\n\n\n\n\n\n\n"
    )
    expected_string = (
        "select 1\n;\n\n\n"
        "select\n    1,\n\n    2\n;\n"
        "{% macro foo() %}\n\n    foo\n{% endmacro %}\n"
    )
    raw_query = formatter.mode.dialect.initialize_analyzer(
        formatter.mode.line_length
    ).parse_query(source_string)
    new_lines = formatter._remove_extra_blank_lines(raw_query.lines)
    assert "".join([str(line) for line in new_lines]) == expected_string



from functools import partial

from sqlfmt import actions
from sqlfmt.rule import Rule
from sqlfmt.rules.common import ALTER_WAREHOUSE, CREATE_WAREHOUSE, group
from sqlfmt.rules.core import CORE
from sqlfmt.tokens import TokenType

WAREHOUSE = [
    *CORE,
    Rule(
        name="unterm_keyword",
        priority=1300,
        pattern=group(
            CREATE_WAREHOUSE,
            ALTER_WAREHOUSE,
            # objectProperties
            r"(with\s+|(un)?set\s+)?"
            + group(
                r"warehouse_type",
                r"warehouse_size",
                r"max_cluster_count",
                r"min_cluster_count",
                r"scaling_policy",
                r"auto_suspend",
                r"auto_resume",
                r"initially_suspended",
                r"resource_monitor",
                r"comment",
                r"enable_query_acceleration",
                r"query_acceleration_max_scale_factor",
                r"tag",
            ),
            # objectParams
            r"(set\s+)?"
            + group(
                r"max_concurrency_level",
                r"statement_queued_timeout_in_seconds",
                r"statement_timeout_in_seconds",
            ),
            # alter
            r"suspend",
            r"resume(\s+if\s+suspended)?",
            r"abort\s+all\s+queries",
            r"rename\s+to",
        )
        + group(r"\W", r"$"),
        action=partial(
            actions.handle_reserved_keyword,
            action=partial(
                actions.add_node_to_buffer, token_type=TokenType.UNTERM_KEYWORD
            ),
        ),
    ),
]



from typing import List

import pytest

from sqlfmt.comment import Comment
from sqlfmt.node import Node
from sqlfmt.node_manager import NodeManager
from sqlfmt.tokens import Token, TokenType


@pytest.fixture
def short_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="-- short comment", spos=0, epos=16
    )
    comment = Comment(t, is_standalone=False, previous_node=None)
    return comment


@pytest.fixture
def short_mysql_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="# short comment", spos=0, epos=15
    )
    comment = Comment(t, is_standalone=False, previous_node=None)
    return comment


@pytest.fixture
def short_js_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="// short comment", spos=0, epos=16
    )
    comment = Comment(t, is_standalone=False, previous_node=None)
    return comment


@pytest.fixture
def nospace_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="--short comment", spos=0, epos=15
    )
    comment = Comment(t, is_standalone=False, previous_node=None)
    return comment


@pytest.fixture
def standalone_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="-- short comment", spos=0, epos=16
    )
    comment = Comment(t, is_standalone=True, previous_node=None)
    return comment


@pytest.fixture
def multiline_comment() -> Comment:
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="/*\ncomment\n*/", spos=0, epos=15
    )
    comment = Comment(t, is_standalone=True, previous_node=None)
    return comment


@pytest.fixture
def datbricks_query_hint_comment() -> Comment:
    n = Node(
        Token(
            type=TokenType.UNTERM_KEYWORD,
            prefix="",
            token="select",
            spos=0,
            epos=6,
        ),
        previous_node=None,
        prefix="",
        value="select",
        open_brackets=[],
        open_jinja_blocks=[],
        formatting_disabled=[],
    )
    t = Token(
        type=TokenType.COMMENT, prefix=" ", token="/*+ hint here */", spos=6, epos=23
    )
    comment = Comment(t, is_standalone=False, previous_node=n)
    return comment


@pytest.fixture
def fmt_disabled_comment() -> Comment:
    t = Token(type=TokenType.FMT_OFF, prefix="", token="--fmt: off", spos=0, epos=10)
    mgr = NodeManager(case_sensitive_names=False)
    node = mgr.create_node(t, None)
    ct = Token(
        type=TokenType.COMMENT,
        prefix="          ",
        token="--   comment",
        spos=11,
        epos=33,
    )
    comment = Comment(ct, is_standalone=True, previous_node=node)
    return comment


def test_get_marker(
    short_comment: Comment,
    short_mysql_comment: Comment,
    nospace_comment: Comment,
    short_js_comment: Comment,
    datbricks_query_hint_comment: Comment,
) -> None:
    assert short_comment._get_marker() == ("--", 3)
    assert short_mysql_comment._get_marker() == ("#", 2)
    assert short_js_comment._get_marker() == ("//", 3)
    assert nospace_comment._get_marker() == ("--", 2)
    assert datbricks_query_hint_comment._get_marker() == ("/*", 2)


def test_comment_parts(
    short_comment: Comment,
    short_mysql_comment: Comment,
    nospace_comment: Comment,
    short_js_comment: Comment,
) -> None:
    assert short_comment._comment_parts() == ("--", "short comment")
    assert short_mysql_comment._comment_parts() == ("#", "short comment")
    assert short_js_comment._comment_parts() == ("--", "short comment")
    assert nospace_comment._comment_parts() == ("--", "short comment")


def test_str_len(
    short_comment: Comment,
    short_mysql_comment: Comment,
    nospace_comment: Comment,
    short_js_comment: Comment,
) -> None:
    assert str(short_comment) == short_comment.token.token
    assert str(short_mysql_comment) == short_mysql_comment.token.token
    assert str(short_js_comment) == f"-- {short_js_comment.body}"
    assert str(nospace_comment) == str(short_comment)

    assert len(short_comment) == 16
    assert len(short_mysql_comment) == 15
    assert len(short_js_comment) == 16
    assert len(nospace_comment) == 16


def test_render_inline(
    short_comment: Comment,
    nospace_comment: Comment,
    short_js_comment: Comment,
) -> None:
    expected = "  -- short comment"
    assert short_comment.render_inline() == expected
    assert short_js_comment.render_inline() == expected
    assert nospace_comment.render_inline() == expected


@pytest.mark.parametrize(
    "prefix",
    [
        "",
        " " * 4,
    ],
)
def test_render_standalone(short_comment: Comment, prefix: str) -> None:
    assert (
        short_comment.render_standalone(max_length=88, prefix=prefix)
        == prefix + str(short_comment) + "\n"
    )
    wrapped_comment = short_comment.render_standalone(max_length=14, prefix=prefix)
    lines = wrapped_comment.splitlines(keepends=True)
    assert lines[0] == prefix + "-- short\n"
    assert lines[1] == prefix + "-- comment\n"


def test_render_standalone_formatting_disabled(fmt_disabled_comment: Comment) -> None:
    assert (
        fmt_disabled_comment.render_standalone(max_length=88, prefix="")
        == f"{fmt_disabled_comment.token.prefix}{fmt_disabled_comment.token.token}\n"
    )


def test_render_standalone_wrap_strip_whitespace() -> None:
    txt = "-- foo" + " " * 100 + "bar"
    t = Token(type=TokenType.COMMENT, prefix="", token=txt, spos=0, epos=len(txt))
    comment = Comment(t, is_standalone=True, previous_node=None)
    assert comment.render_standalone(max_length=88, prefix="") == "-- foo\n-- bar\n"


def test_render_multiline(multiline_comment: Comment) -> None:
    assert str(multiline_comment) == "/*\ncomment\n*/"
    assert (
        multiline_comment.render_standalone(max_length=88, prefix="")
        == f"{multiline_comment}\n"
    )


@pytest.mark.parametrize(
    "text,expected_splits",
    [
        ("asdf", ["asdf"]),
        ("asdfqwerzxcv", ["asdfqwerzxcv"]),
        ("asdf qwer", ["asdf qwer"]),
        ("asdf qwer zxcv", ["asdf qwer", "zxcv"]),
        ("asdf qwer zxcv uiophjkl vbnm", ["asdf qwer", "zxcv", "uiophjkl", "vbnm"]),
    ],
)
def test_split_before(text: str, expected_splits: List[str]) -> None:
    result = list(Comment._split_before(text, max_length=10))
    assert result == expected_splits


def test_empty_comment() -> None:
    t = Token(type=TokenType.COMMENT, prefix=" ", token="-- ", spos=0, epos=3)
    comment = Comment(t, is_standalone=True, previous_node=None)
    assert str(comment) == "--"


def test_is_inline(
    short_comment: Comment, standalone_comment: Comment, multiline_comment: Comment
) -> None:
    assert short_comment.is_inline
    assert not (standalone_comment.is_inline)
    assert not (multiline_comment.is_inline)


def test_no_wrap_long_jinja_comments() -> None:
    comment_str = "{# " + ("comment " * 20) + "#}"
    t = Token(
        type=TokenType.COMMENT,
        prefix=" ",
        token=comment_str,
        spos=0,
        epos=len(comment_str),
    )
    comment = Comment(t, is_standalone=True, previous_node=None)
    rendered = comment.render_standalone(88, "")

    assert rendered == comment_str + "\n"


def test_no_add_space_databricks_hint(datbricks_query_hint_comment: Comment) -> None:
    assert str(datbricks_query_hint_comment) == datbricks_query_hint_comment.token.token



from pathlib import Path
from typing import List, Optional, Union

import click

from sqlfmt import api
from sqlfmt.config import load_config_file
from sqlfmt.mode import Mode


@click.command()
@click.version_option(package_name="shandy-sqlfmt")
@click.option(
    "--check",
    envvar="SQLFMT_CHECK",
    is_flag=True,
    help=(
        "Fail with an exit code of 1 if source files are not formatted to spec. "
        "Do not write formatted queries to files."
    ),
)
@click.option(
    "--diff",
    envvar="SQLFMT_DIFF",
    is_flag=True,
    help=(
        "Print a diff of any formatting changes to stdout. Fails like --check "
        "on any changes. Do not write formatted queries to files."
    ),
)
@click.option(
    "--exclude",
    envvar="SQLFMT_EXCLUDE",
    multiple=True,
    help=(
        "A string that is passed to glob.glob as a pathname; any matching files "
        "returned by glob will be excluded from FILES and not formatted. Note "
        "that glob is relative to the current working directory when sqlfmt is "
        "called. To exclude multiple globs, repeat the --exclude option."
    ),
)
@click.option(
    "--encoding",
    envvar="SQLFMT_ENCODING",
    help=(
        "The encoding to use when reading and writing .sql files. Defaults "
        "to utf-8. Set to 'inherit' to read the system default encoding. utf "
        "encodings will detect and preserve the BOM if one is present."
    ),
)
@click.option(
    "--fast/--safe",
    envvar="SQLFMT_FAST",
    default=False,
    help=(
        "By default, sqlfmt re-processes the output it produces in "
        "order to run a safety check and ensure that all tokens from "
        "the input are present in the output. This can add 15-20% to "
        "the processing time for new files. To disable this safety "
        "check, use the --fast option. To force the safety check, "
        "use --safe."
    ),
)
@click.option(
    "--single-process",
    envvar="SQLFMT_SINGLE_PROCESS",
    is_flag=True,
    help=(
        "Run sqlfmt in a single process, even when formatting multiple "
        "files. If not set, defaults to multiprocessing using as many "
        "cores as possible."
    ),
)
@click.option(
    "-k",
    "--reset-cache",
    envvar="SQLFMT_RESET_CACHE",
    is_flag=True,
    help=(
        "Clear the sqlfmt cache before running, effectively forcing sqlfmt "
        "to operate on every file. Will slow down runs."
    ),
)
@click.option(
    "--no-jinjafmt",
    envvar="SQLFMT_NO_JINJAFMT",
    is_flag=True,
    help=(
        "Do not format jinja tags (the code between the curlies). Only necessary "
        "to specify this flag if sqlfmt was installed with the jinjafmt extra, "
        "or if black was already available in this environment."
    ),
)
@click.option(
    "-l",
    "--line-length",
    envvar="SQLFMT_LINE_LENGTH",
    default=88,
    type=int,
    help=("The maximum line length allowed in output files. Default is 88."),
)
@click.option(
    "-v",
    "--verbose",
    envvar="SQLFMT_VERBOSE",
    is_flag=True,
    help=("Prints more information to stderr."),
)
@click.option(
    "-q",
    "--quiet",
    envvar="SQLFMT_QUIET",
    is_flag=True,
    help=(
        "Stop emitting all non-critical output. Error messages will still be "
        "emitted (which can silenced by 2>/dev/null)."
    ),
)
@click.option(
    "--no-progressbar",
    envvar="SQLFMT_NO_PROGRESSBAR",
    is_flag=True,
    help=("Never prints a progressbar to stderr."),
)
@click.option(
    "--no-color",
    envvar="SQLFMT_NO_COLOR",
    is_flag=True,
    help=(
        "Removes color codes from all output, including diffs. "
        "Alternatively, set the NO_COLOR environment variable. "
        "See https://no-color.org/ for more details."
    ),
)
@click.option(
    "--force-color",
    envvar="SQLFMT_FORCE_COLOR",
    is_flag=True,
    help=(
        "sqlfmt output is colorized by default. However, if you have "
        "the NO_COLOR env var set, and still want sqlfmt to colorize "
        "output, you can use --force-color to override the env var."
    ),
)
@click.option(
    "-d",
    "--dialect",
    "dialect_name",
    envvar="SQLFMT_DIALECT",
    type=click.Choice(["polyglot", "clickhouse"], case_sensitive=False),
    default="polyglot",
    help=(
        "The SQL dialect for the target files. Nearly all dialects are supported "
        "by the default polyglot dialect. Select the ClickHouse dialect to respect "
        "case sensitivity in function, field, and alias names."
    ),
)
@click.option(
    "--config",
    "config_path",
    envvar="SQLFMT_CONFIG",
    type=click.Path(
        exists=True, dir_okay=False, allow_dash=False, resolve_path=True, path_type=Path
    ),
    help=(
        "A path to a `pyproject.toml` file. Options passed at the command line will "
        "override settings in this file."
    ),
)
@click.argument(
    "files",
    nargs=-1,
    type=click.Path(exists=True, allow_dash=True, resolve_path=True, path_type=Path),
)
@click.pass_context
def sqlfmt(
    ctx: click.Context,
    files: List[Path],
    config_path: Optional[Path] = None,
    **kwargs: Union[bool, int, List[str], str],
) -> None:
    """
    sqlfmt formats your dbt SQL files so you don't have to.

    The FILES argument can be one or many paths to sql files (or directories),
    or use "-" to use stdin.

    Exit codes: 0 indicates success, 1 indicates failed check,
    2 indicates a handled exception caused by errors in one or more user code files.

    https://sqlfmt.com for documentation and more information.
    """
    if files:
        config = load_config_file(files, config_path)
        non_default_options = {
            k: v
            for k, v in kwargs.items()
            if ctx.get_parameter_source(k).name != "DEFAULT"  # type: ignore
        }
        config.update(non_default_options)
        mode = Mode(**config)  # type: ignore

        matched_files = api.get_matching_paths(files, mode=mode)
        progress_bar, progress_callback = api.initialize_progress_bar(
            total=len(matched_files), mode=mode
        )

        report = api.run(files=matched_files, mode=mode, callback=progress_callback)

        progress_bar.close()
        report.display_report()

        if report.number_errored > 0:
            exit_code = 2
        elif (mode.check or mode.diff) and report.number_changed > 0:
            exit_code = 1
        else:
            exit_code = 0
    else:
        show_welcome_message()
        exit_code = 0

    ctx.exit(exit_code)


def show_welcome_message() -> None:
    """
    Prints a nice welcome message for new users who might accidentally
    enter `$ sqlfmt` without any arguments
    """
    from sqlfmt.report import display_output, style_output

    art = r"""
               _  __           _
              | |/ _|         | |
     ___  __ _| | |_ _ __ ___ | |_
    / __|/ _` | |  _| '_ ` _ \| __|
    \__ \ (_| | | | | | | | | | |_
    |___/\__, |_|_| |_| |_| |_|\__|
            | |
            |_|"""
    display_output(msg=art)
    message = """
    sqlfmt formats your dbt SQL files so you don't have to.
    For more information, visit https://sqlfmt.com

    To get started, try:
    """
    display_output(msg=message)
    commands = [
        (
            "sqlfmt .",
            "format all files nested in the current dir (note the '.')",
        ),
        (
            "sqlfmt path/to/file.sql",
            "format file.sql only",
        ),
        (
            "sqlfmt . --check",
            "check formatting of all files, exit with code 1 on changes",
        ),
        (
            "sqlfmt . --diff",
            "print diff resulting from formatting all files",
        ),
        (
            "sqlfmt -",
            "format text received through stdin, write result to stdout",
        ),
        (
            "sqlfmt --help",
            "show more options and other usage information",
        ),
    ]
    margin = max([len(cmd) for cmd, _ in commands])
    for cmd, desc in commands:
        styled_cmd = style_output(
            msg=f"{cmd}{' ' * (margin - len(cmd))}",
            fg="white",
            bg="bright_black",
            bold=True,
        )
        display_output(msg=f"    {styled_cmd} {desc}")
    display_output(msg="\n")
